import numpy as np
from commonroad.common.util import Interval, AngleInterval
from commonroad.geometry.shape import Rectangle, Polygon, Circle, ShapeGroup
from commonroad.planning.goal import GoalRegion
from commonroad.planning.planning_problem import PlanningProblem
from commonroad.scenario.lanelet import Lanelet
from commonroad.scenario.scenario import Scenario, ScenarioID
from commonroad.scenario.trajectory import State, CustomState
from commonroad_route_planner.route_planner import RoutePlanner

from commonroad_rl.gym_commonroad.observation import GoalObservation
from commonroad_rl.gym_commonroad.utils.navigator import Navigator
from commonroad_rl.tests.common.marker import *

dummy_time_step = Interval(0.0, 0.0)


@pytest.mark.parametrize(
    ("ego_orientation", "expected_output"),
    [
        (-np.pi / 2., np.pi / 4.),
        (np.pi / 2., -np.pi / 4.),
        (-0.5, 0.),
        (0., 0.),
        (np.pi * 15. / 8., 0.),
        (0.5, 0.),
        (np.pi, 3 * np.pi / 4.),
    ],
)
@unit_test
@functional
def test_get_goal_orientation_distance(ego_orientation, expected_output):
    """Tests GoalObservation._get_goal_orientation_distance"""
    configs = {"goal_configs": {}}

    dummy_state = {
        "velocity": 0.0,
        "yaw_rate": 0.0,
        "slip_angle": 0.0,
        "time_step": 0.0,
        "position": np.array([0.0, 0.0])
    }
    ego_state = CustomState(**dummy_state, orientation=ego_orientation)
    goal_state = CustomState(time_step=dummy_time_step,
                       position=Rectangle(length=2.0, width=2.0, center=np.array([5.0, 0.0])),
                       orientation=AngleInterval(-np.pi / 4, np.pi / 4))

    goal_obs = GoalObservation(configs)

    min_goal_orientation_distance = goal_obs._get_goal_orientation_distance(ego_state.orientation,
                                                                            GoalRegion([goal_state]))

    assert np.isclose(min_goal_orientation_distance, expected_output)


# [10, 20]
@pytest.mark.parametrize(
    ("ego_time_step", "expected_output"),
    [
        (5, -5),
        (10, 0),
        (15, 0),
        (20, 0),
        (30, 10),
    ],
)
@unit_test
@functional
def test_get_goal_time_distance(ego_time_step, expected_output):
    """Tests GoalObservation._get_goal_time_distance"""

    dummy_state = {
        "velocity": 0.0,
        "yaw_rate": 0.0,
        "slip_angle": 0.0,
        "orientation": 0.0,
        "position": np.array([0.0, 0.0])
    }
    ego_state = CustomState(**dummy_state, time_step=ego_time_step)
    goal_state = CustomState(time_step=Interval(10, 20),
                       position=Rectangle(length=2.0, width=2.0, center=np.array([5.0, 0.0])),
                       orientation=AngleInterval(-np.pi / 2, np.pi / 2))

    min_goal_time_distance = GoalObservation._get_goal_time_distance(ego_state.time_step, GoalRegion([goal_state]))

    assert np.isclose(min_goal_time_distance, expected_output)


@pytest.mark.parametrize(
    ("ego_velocity", "expected_output"),
    [
        (5., -5.),
        (10., 0.),
        (15., 0.),
        (20., 0.),
        (30., 10.),
    ],
)
@unit_test
@functional
def test_get_goal_velocity_distance(ego_velocity, expected_output):
    """Tests GoalObservation._get_goal_velocity_distance"""

    dummy_state = {
        "time_step": dummy_time_step,
        "yaw_rate": 0.0,
        "slip_angle": 0.0,
        "orientation": 0.0,
        "position": np.array([0.0, 0.0])
    }
    ego_state = CustomState(**dummy_state, velocity=ego_velocity)
    goal_state = CustomState(time_step=Interval(10, 20),
                       position=Rectangle(length=2.0, width=2.0, center=np.array([5.0, 0.0])),
                       orientation=AngleInterval(-np.pi / 2, np.pi / 2),
                       velocity=Interval(10., 20.))

    min_goal_velocity_distance = GoalObservation._get_goal_velocity_distance(ego_state.velocity,
                                                                             GoalRegion([goal_state]))

    assert np.isclose(min_goal_velocity_distance, expected_output)


@pytest.mark.parametrize(
    ("prev_advances", "lat_long_distance", "expected_output"),
    [
        ((5, 5), (2, 3), (3, 2)),
        ((5, 5), (7, 9), (-2, -4)),
        ((-13, -3), (-5, 3), (8, 0)),
        ((-3, -1), (-4, -2), (-1, -1)),
        ((-2, -2), (0, 4), (2, -2))
    ]
)
@unit_test
@functional
def test_get_long_lat_distance_advance_to_goal(prev_advances, lat_long_distance, expected_output):
    """Tests GoalObservation._get_long_lat_distance_advance_to_goal"""
    configs = {"goal_configs": {'observe_distance_goal_long': True,
                                'observe_distance_goal_lat': True}}

    dummy_state = {
        "time_step": Interval(1.0, 1.0),
        "yaw_rate": 0.0,
        "slip_angle": 0.0,
        "orientation": 0.0,
        "position": np.array([0.0, 0.0])
    }
    ego_state = CustomState(**dummy_state)

    goal_obs = GoalObservation(configs)
    goal_obs.observation_history_dict['distance_goal_long'] = prev_advances[0]
    goal_obs.observation_history_dict['distance_goal_lat'] = prev_advances[1]

    advance = goal_obs._get_long_lat_distance_advance_to_goal(*lat_long_distance)

    assert np.isclose(advance[0], expected_output[0])
    assert np.isclose(advance[1], expected_output[1])

    ego_state.time_step = 1
    goal_obs.observe_distance_goal_long = False
    goal_obs.observe_distance_goal_lat = False

    advance = goal_obs._get_long_lat_distance_advance_to_goal(*lat_long_distance)
    assert np.isclose(advance[0], 0)
    assert np.isclose(advance[1], 0)


@pytest.mark.parametrize(
    ('ego_position', 'goal_region', 'desired_distance'),
    [
        (
                np.array([0, 0]),
                GoalRegion([CustomState(time_step=Interval(1, 1),
                                  position=Rectangle(length=2.0, width=2.0, center=np.array([0.0, 0.0])),
                                  orientation=AngleInterval(-np.pi / 2, np.pi / 2),
                                  velocity=Interval(0., 0.))]),
                0.0
        ),
        (
                np.array([0, 0]),
                GoalRegion([CustomState(time_step=Interval(1, 1),
                                  position=Rectangle(length=2.0, width=2.0, center=np.array([1.0, 1.0])),
                                  orientation=AngleInterval(-np.pi / 2, np.pi / 2),
                                  velocity=Interval(0., 0.))]),
                np.sqrt(2.0)
        ),
        (
                np.array([0, 0]),
                GoalRegion([CustomState(time_step=Interval(1, 1),
                                  position=Polygon(np.array([[0, 0], [0, 2], [2, 2], [2, 0]])),
                                  orientation=AngleInterval(-np.pi / 2, np.pi / 2),
                                  velocity=Interval(0., 0.))]),
                np.sqrt(2.0)
        ),
        (
                np.array([0, 0]),
                GoalRegion([CustomState(time_step=Interval(1, 1),
                                  position=Circle(1, np.array([1, 1])),
                                  orientation=AngleInterval(-np.pi / 2, np.pi / 2),
                                  velocity=Interval(0., 0.))]),
                np.sqrt(2.0)
        ),
        (
                np.array([0, 0]), GoalRegion([
                    CustomState(time_step=Interval(1, 1),
                          position=ShapeGroup([
                              Polygon(np.array([[-7.3275, 12.5257],
                                                [-7.5254, 9.1777],
                                                [-11.278, 9.1652],
                                                [-15.0305, 9.1526],
                                                [-15.1272, 12.6073],
                                                [-11.2273, 12.5665]])),
                              Polygon(np.array([[-55.7567, 6.6837],
                                                [-54.5595, 3.548],
                                                [-60.9422, 1.1307],
                                                [-67.3249, -1.2866],
                                                [-68.7141, 1.7041],
                                                [-62.2354, 4.1939]])),
                              Polygon(np.array([[-15.1272, 12.6073],
                                                [-15.0305, 9.1526],
                                                [-21.3058, 9.0022],
                                                [-27.581, 8.8518],
                                                [-27.8613, 12.1],
                                                [-21.4942, 12.3536]])),
                              Polygon(np.array([
                                  [-27.8613, 12.1],
                                  [-27.581, 8.8518],
                                  [-34.3544, 8.3423],
                                  [-41.5165, 7.3883],
                                  [-48.0403, 5.7701],
                                  [-54.5595, 3.548],
                                  [-55.7567, 6.6837],
                                  [-48.7239, 8.8909],
                                  [-41.8512, 10.351],
                                  [-34.9102, 11.5653]
                              ]))
                          ]),
                          orientation=AngleInterval(-np.pi / 2, np.pi / 2),
                          velocity=Interval(0., 0.))
                ]),
                34.9101481
        )
    ]
)
@unit_test
@functional
def test_get_goal_euclidean_distance(ego_position: np.ndarray, goal_region: GoalRegion, desired_distance: float):
    actual_distance = GoalObservation._get_goal_euclidean_distance(ego_position, goal_region)
    assert np.isclose(actual_distance, desired_distance)

@pytest.mark.parametrize(
    ("shape_group", "actual_center"),
    [
        (
                ShapeGroup([
                    Polygon(np.array([[-7.3275, 12.5257],
                                      [-7.5254, 9.1777],
                                      [-11.278, 9.1652],
                                      [-15.0305, 9.1526],
                                      [-15.1272, 12.6073],
                                      [-11.2273, 12.5665]])),
                    Polygon(np.array([[-55.7567, 6.6837],
                                      [-54.5595, 3.548],
                                      [-60.9422, 1.1307],
                                      [-67.3249, -1.2866],
                                      [-68.7141, 1.7041],
                                      [-62.2354, 4.1939]])),
                    Polygon(np.array([[-15.1272, 12.6073],
                                      [-15.0305, 9.1526],
                                      [-21.3058, 9.0022],
                                      [-27.581, 8.8518],
                                      [-27.8613, 12.1],
                                      [-21.4942, 12.3536]])),
                    Polygon(np.array([
                        [-27.8613, 12.1],
                        [-27.581, 8.8518],
                        [-34.3544, 8.3423],
                        [-41.5165, 7.3883],
                        [-48.0403, 5.7701],
                        [-54.5595, 3.548],
                        [-55.7567, 6.6837],
                        [-48.7239, 8.8909],
                        [-41.8512, 10.351],
                        [-34.9102, 11.5653]
                    ]))
                ]),
                np.array([-33.93858901,   8.17866797])
        ),
        (
                ShapeGroup(shapes=[
                    Circle(4,np.array([0,0])),
                    Circle(100,np.array([-1,1])),
                    Polygon(np.array([
                        [4,4],
                        [8,4],
                        [8,8],
                        [6,6],
                        [4,8],
                    ]))
                ]),
                np.array([5/3, 6.555555/3])
        )
    ]
)
@unit_test
@functional
def test_convert_shape_group_to_center(shape_group: ShapeGroup, actual_center: np.ndarray):
    center = GoalObservation._convert_shape_group_to_center(shape_group)
    assert all(np.isclose(center, actual_center))

@pytest.mark.parametrize(
    ('ego_state', 'goal_region', 'is_reached'),
    [
        (
                CustomState(**{
                    "time_step": 1,
                    "yaw_rate": 0.0,
                    "slip_angle": 0.0,
                    "orientation": 0.0,
                    "position": np.array([0.0, 0.0]),
                    "velocity": 0
                }),
                GoalRegion([CustomState(time_step=Interval(1, 1),
                                  position=Rectangle(length=2.0, width=2.0, center=np.array([0.0, 0.0])),
                                  orientation=AngleInterval(-np.pi / 2, np.pi / 2),
                                  velocity=Interval(0., 0.))]),
                True
        ),
        (
                CustomState(**{
                    "time_step": 5,
                    "yaw_rate": 0.0,
                    "slip_angle": 0.0,
                    "orientation": 0.0,
                    "position": np.array([0.0, 0.0]),
                    "velocity": 0
                }),
                GoalRegion([CustomState(time_step=Interval(1, 1),
                                  position=Rectangle(length=2.0, width=2.0, center=np.array([0.0, 0.0])),
                                  orientation=AngleInterval(-np.pi / 2, np.pi / 2),
                                  velocity=Interval(0., 0.))]),
                False
        ),
        (
                CustomState(**{
                    "time_step": 1,
                    "yaw_rate": 0.0,
                    "slip_angle": 0.0,
                    "orientation": 0.0,
                    "position": np.array([4.0, 0.0]),
                    "velocity": 0
                }),
                GoalRegion([CustomState(time_step=Interval(1, 1),
                                  position=Rectangle(length=2.0, width=2.0, center=np.array([0.0, 0.0])),
                                  orientation=AngleInterval(-np.pi / 2, np.pi / 2),
                                  velocity=Interval(0., 0.))]),
                False
        ),
        (
                CustomState(**{
                    "time_step": 1,
                    "yaw_rate": 0.0,
                    "slip_angle": 0.0,
                    "orientation": 3 * np.pi / 2,
                    "position": np.array([0.0, 0.0]),
                    "velocity": 0.
                }),
                GoalRegion([CustomState(time_step=Interval(1, 1),
                                  position=Rectangle(length=2.0, width=2.0, center=np.array([0.0, 0.0])),
                                  orientation=AngleInterval(-np.pi / 2, np.pi / 2),
                                  velocity=Interval(0., 0.))]),
                True
        )
    ]
)
@functional
@unit_test
def test_check_goal_reached(ego_state: State, goal_region: GoalRegion, is_reached: bool):
    """unittest for the GoalObservation._check_goal_reached method"""
    assert GoalObservation._check_goal_reached(goal_region, ego_state) == is_reached


@pytest.mark.parametrize(
    ('ego_state', 'goal_region', 'time_out'),
    [
        (
                CustomState(**{
                    "time_step": 1
                }),
                GoalRegion([CustomState(time_step=Interval(1, 1))]),
                True
        ),
        (
                CustomState(**{
                    "time_step": 3
                }),
                GoalRegion([CustomState(time_step=Interval(1, 1))]),
                True
        ),
        (
                CustomState(**{
                    "time_step": 0
                }),
                GoalRegion([CustomState(time_step=Interval(1, 1))]),
                False
        )
    ]
)
@functional
@unit_test
def test_check_is_time_out(ego_state: State, goal_region: GoalRegion, time_out: bool):
    """unittest for the GoalObservation._check_is_time_out method"""

    assert GoalObservation._check_is_time_out(ego_state, goal_region, False) == time_out
    assert GoalObservation._check_is_time_out(ego_state, goal_region, True) == False


@pytest.mark.parametrize(
    ("ego_position", "expected_output"),
    [
        (np.array([0.0, 0.0]), (4.0, 0.0)),
        (np.array([4.0, 0.0]), (0.0, 0.0)),
        (np.array([5.0, 0.0]), (0.0, 0.0)),
        (np.array([6.0, 0.0]), (0.0, 0.0)),
        (np.array([10.0, 0.0]), (-4.0, 0.0)),
        (np.array([0.0, 1.0]), (4.0, 0.0)),
        (np.array([4.0, 1.0]), (0.0, 0.0)),
        (np.array([5.0, 1.0]), (0.0, 0.0)),
        (np.array([6.0, 1.0]), (0.0, 0.0)),
        (np.array([10.0, 1.0]), (-4.0, 0.0)),
        (np.array([0.0, 2.0]), (4.0, -1.0)),
        (np.array([4.0, 2.0]), (0.0, -1.0)),
        (np.array([5.0, 2.0]), (0.0, -1.0)),
        (np.array([6.0, 2.0]), (0.0, -1.0)),
        (np.array([10.0, 2.0]), (-4.0, -1.0)),
        (np.array([0.0, -1.0]), (4.0, 0.0)),
        (np.array([4.0, -1.0]), (0.0, 0.0)),
        (np.array([5.0, -1.0]), (0.0, 0.0)),
        (np.array([6.0, -1.0]), (0.0, 0.0)),
        (np.array([10.0, -1.0]), (-4.0, 0.0)),
        (np.array([0.0, -2.0]), (4.0, 1.0)),
        (np.array([4.0, -2.0]), (0.0, 1.0)),
        (np.array([5.0, -2.0]), (0.0, 1.0)),
        (np.array([6.0, -2.0]), (0.0, 1.0)),
        (np.array([10.0, -2.0]), (-4.0, 1.0)),
    ],
)
@unit_test
@functional
def test_get_long_lat_distance_to_goal(ego_position, expected_output):
    """Tests the GoalObservation._get_long_lat_distance_to_goal function"""

    dummy_state = {
        "velocity": 0.0,
        "orientation": 0.0,
        "yaw_rate": 0.0,
        "slip_angle": 0.0,
        "time_step": 0.0,
    }
    ego_state = CustomState(**dummy_state, position=ego_position)
    goal_state = CustomState(time_step=dummy_time_step,
                       position=Rectangle(length=2.0, width=2.0, center=np.array([5.0, 0.0])))
    planning_problem = PlanningProblem(planning_problem_id=0, initial_state=ego_state,
                                       goal_region=GoalRegion([goal_state]))

    lanelet = Lanelet(lanelet_id=0,
                      left_vertices=np.array([[0.0, 3.0], [10.0, 3.0]]),
                      center_vertices=np.array([[0.0, 0.0], [10.0, 0.0]]),
                      right_vertices=np.array([[0.0, -3.0], [10.0, -3.0]]))

    scenario = Scenario(dt=0.1, scenario_id=ScenarioID("DEU_TEST-1_1_T-1"))
    scenario.lanelet_network.add_lanelet(lanelet)

    route_planner = RoutePlanner(
        scenario,
        planning_problem,
        backend=RoutePlanner.Backend.NETWORKX_REVERSED,
        log_to_console=False,
    )

    route_candidates = route_planner.plan_routes()
    route = route_candidates.retrieve_best_route_by_orientation()

    navigator = Navigator(route)

    min_distance_long, min_distance_lat = GoalObservation.get_long_lat_distance_to_goal(ego_state.position, navigator)
    assert np.isclose(min_distance_long, expected_output[0]) and np.isclose(min_distance_lat, expected_output[1])



import matplotlib.pyplot as plt
import pickle

with open("continuous_actions.pkl", "rb") as f:
    continuous_actions = pickle.load(f)
with open("discrete_actions.pkl", "rb") as f:
    discrete_actions = pickle.load(f)
# continuous_actions = np.load("continuous_actions.npz", allow_pickle=True)
# discrete_actions = np.load("discrete_actions.npz", allow_pickle=True)

discrete_accelerations = discrete_actions["accelerations"]
discrete_jerks = discrete_actions["jerks"]
continuous_acceleraitons = continuous_actions["accelerations"]
continuous_jerks = continuous_actions["accelerations"]

for key in discrete_accelerations.keys():
    plt.figure()
    # accelerations
    plt.subplot(2, 2, 1)
    plt.title("accleration-x")
    plt.plot(discrete_accelerations[key][:, 0], color="blue", label="discrete")
    plt.plot(continuous_acceleraitons[key][:, 0], color="red", label="continuous")
    plt.legend()
    plt.subplot(2, 2, 2)
    plt.title("accleration-y")
    plt.plot(discrete_accelerations[key][:, 1], color="blue", label="discrete")
    plt.plot(continuous_acceleraitons[key][:, 1], color="red", label="continuous")
    # plt.legend()
    # jerks
    plt.subplot(2, 2, 3)
    plt.title("jerk-x")
    plt.plot(discrete_jerks[key][:, 0], color="blue", label="discrete")
    plt.plot(continuous_jerks[key][:, 0], color="red", label="continuous")
    # plt.legend()
    plt.subplot(2, 2, 4)
    plt.title("jerk-y")
    plt.plot(discrete_jerks[key][:, 1], color="blue", label="discrete")
    plt.plot(continuous_jerks[key][:, 1], color="red", label="continuous")
    # plt.legend()
    plt.tight_layout()
    plt.savefig(f"{key}.png")


"""
Module for training an agent using stable baselines
"""
import os

os.environ["KMP_WARNINGS"] = "off"
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"
os.environ["KMP_AFFINITY"] = "none"
import sys
import logging

logging.getLogger("tensorflow").disabled = True
import gym
import time
import uuid
import copy
import yaml
import difflib
import argparse
import importlib
import numpy as np
from enum import Enum
from pprint import pformat
from stable_baselines.common.base_class import _UnvecWrapper
from stable_baselines.her import HERGoalEnvWrapper
from stable_baselines.gail import ExpertDataset
from stable_baselines import logger
# NUM_PARALLEL_EXEC_UNITS = 4
# os.environ["OMP_NUM_THREADS"] = str(NUM_PARALLEL_EXEC_UNITS)
os.environ["KMP_AFFINITY"] = "none"  # "granularity=fine,verbose,compact,1,0"

SB_LOGGER = logger.Logger(folder=None, output_formats=[logger.HumanOutputFormat(sys.stdout)])
LOGGER = logging.getLogger(__name__)

import commonroad_rl.gym_commonroad
from commonroad_rl.gym_commonroad.constants import ROOT_STR, PATH_PARAMS
from commonroad_rl.utils_run.callbacks import (
    SaveVecNormalizeCallback,
    MultiEnvsEvalCallback,
)
from commonroad_rl.utils_run.hyperparams_opt import optimize_hyperparams
from commonroad_rl.utils_run.observation_configs_opt import optimize_observation_configs
from commonroad_rl.utils_run.reward_configs_opt import optimize_reward_configs
from commonroad_rl.utils_run.noise import LinearNormalActionNoise
from commonroad_rl.utils_run.utils import (
    StoreDict,
    linear_schedule,
    get_wrapper_class,
    get_latest_run_id,
    make_env,
    ALGOS,
)

# numpy warnings because of tensorflow
import warnings

warnings.filterwarnings(action="ignore", category=FutureWarning, module="tensorflow")
warnings.filterwarnings(action="ignore", category=UserWarning, module="gym")
warnings.simplefilter(action="ignore", category=FutureWarning)
import tensorflow as tf

tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
# Optional dependencies
try:
    import mpi4py
    from mpi4py import MPI
except ImportError:
    mpi4py = None

from stable_baselines.common import set_global_seeds
from stable_baselines.common.cmd_util import make_atari_env
from stable_baselines.common.vec_env import (
    VecFrameStack,
    VecNormalize,
    DummyVecEnv,
    SubprocVecEnv,
    VecEnv,
)
from stable_baselines.common.noise import (
    AdaptiveParamNoiseSpec,
    NormalActionNoise,
    OrnsteinUhlenbeckActionNoise,
)
from stable_baselines.common.schedules import constfn
from stable_baselines.common.callbacks import CheckpointCallback, EvalCallback


class LoggingMode(Enum):
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    ERROR = logging.ERROR


def run_stable_baselines_argsparser():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--env", type=str, default="commonroad-v1", help="environment ID")
    parser.add_argument("-tb", "--tensorboard-log", help="Tensorboard log dir", default="", type=str)
    parser.add_argument("-i", "--trained-agent", type=str, default="",
                        help="Path to a pretrained agent to continue training")
    parser.add_argument("--algo", help="RL Algorithm", default="ppo2", type=str,
                        required=False, choices=list(ALGOS.keys()))
    parser.add_argument("-n", "--n-timesteps", help="Set the number of timesteps", default=int(1e6), type=int)
    parser.add_argument("--log-interval", help="Override log interval (default: -1, no change)", default=-1, type=int)
    parser.add_argument("--eval-freq", default=10000, type=int,
                        help="Evaluate the agent every n steps (if negative, no evaluation)")
    parser.add_argument("--eval_timesteps", help="Number of timesteps to use for evaluation", default=1000, type=int)
    parser.add_argument("--eval_episodes", help="Number of episodes to use for evaluation", default=5, type=int)
    parser.add_argument("--save-freq", default=-1, type=int,
                        help="Save the model every n steps (if negative, no checkpoint)")
    parser.add_argument("-f", "--log-folder", help="Log folder", type=str, default="logs")
    parser.add_argument("--save_path", type=str, default=None)
    parser.add_argument("--seed", help="Random generator seed", type=int, default=0)
    parser.add_argument("--configs-path", type=str, default="",
                        help="Path to file for overwriting environment configurations")
    parser.add_argument("--hyperparams-path", type=str, default="",
                        help="Path to file for overwriting model hyperparameters")
    parser.add_argument("--optimize-observation-configs", action="store_true", default=False,
                        help="Optimize observation configurations")
    parser.add_argument("--optimize-reward-configs", action="store_true", default=False,
                        help="Optimize reward configurations")
    parser.add_argument("--guided", action="store_true", default=False,
                        help="Guided optimization of some reward parameters")
    parser.add_argument("--optimize-hyperparams", action="store_true", default=False,
                        help="Optimize model hyperparameters")
    parser.add_argument("--n-trials", help="Number of trials for optimization", type=int, default=5)
    parser.add_argument("--n-jobs", help="Number of parallel jobs for optimization", type=int, default=1)
    parser.add_argument("--sampler", type=str, default="tpe", choices=["random", "tpe", "skopt"],
                        help="Sampler for optimization")
    parser.add_argument("--pruner", help="Pruner for optimization", type=str,
                        default="median", choices=["halving", "median", "none"])
    parser.add_argument("--logging_mode", default=LoggingMode.INFO, type=LoggingMode, choices=list(LoggingMode))
    parser.add_argument("--gym-packages", type=str, nargs="+", default=[],
                        help="Additional external Gym environment package modules to import (e.g. gym_minigrid)")
    parser.add_argument("-params", "--hyperparams", type=str, nargs="+", action=StoreDict,
                        help="Overwrite model hyperparameters (e.g. learning_rate:0.01 train_freq:10)")
    parser.add_argument("-uuid", "--uuid", choices=["top", "none", "true"], type=str, default="none",
                        help="Ensure that the run has a unique ID")
    parser.add_argument("--env-kwargs", type=str, nargs="+", action=StoreDict,
                        help='Overwrite environment configurations '
                             '(e.g. observe_heading:"True" reward_type:"\'default_reward\'" '
                             'meta_scenario_path:"str(\'./DEU_LocationA-11_11_T-1/meta_scenario\')")')
    parser.add_argument("--n_envs", help="Number of parallel training processes", type=int, default=1)
    parser.add_argument("--info_keywords", type=str, nargs="+", default=(),
                        help="(tuple) extra information to log, from the information return of environment.step, "
                             "see stable_baselines/bench/monitor.py")

    return parser


def construct_logger(logging_mode: int, save_path: str, logger):
    logger.setLevel(logging_mode)
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging_mode)
    file_handler = logging.FileHandler(os.path.join(save_path, "console_copy.txt"))
    file_handler.setLevel(logging_mode)

    formatter = logging.Formatter("[%(levelname)s] %(name)s - %(message)s")
    stream_handler.setFormatter(formatter)
    file_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    logger.addHandler(file_handler)
    logger.debug(f"Logging to console and {save_path + '/console_copy.txt'}")


def create_parameter_schedule(hyperparams: dict):
    # Create learning rate schedules
    for key in ["learning_rate", "cliprange", "cliprange_vf"]:
        if key not in hyperparams:
            continue
        if isinstance(hyperparams[key], str):
            schedule, initial_value = hyperparams[key].split("_")
            initial_value = float(initial_value)
            hyperparams[key] = linear_schedule(initial_value)
        elif isinstance(hyperparams[key], (float, int)):
            # Negative value: ignore (ex: for clipping)
            if hyperparams[key] < 0:
                continue
            hyperparams[key] = constfn(float(hyperparams[key]))
        else:
            raise ValueError(f"Invalid value for {key}: {hyperparams[key]}")
    return hyperparams


def construct_save_path(args):
    log_path = os.path.join(args.log_folder, args.algo)
    if args.uuid == "top":
        return args.log_folder
    elif args.uuid == "true":
        uuid_str = f"_{uuid.uuid4()}"
    else:
        uuid_str = ""
    return os.path.join(log_path, f"{args.env}_{get_latest_run_id(log_path, args.env) + 1}{uuid_str}")


def create_vec_normalized_env(exp_folder: str, env: VecNormalize):
    if isinstance(env, VecNormalize):
        env = env.venv
    if os.path.exists(os.path.join(exp_folder, "vecnormalize.pkl")):
        LOGGER.info(f"Loading running average from {os.path.join(exp_folder, 'vecnormalize.pkl')}")
        return VecNormalize.load(os.path.join(exp_folder, 'vecnormalize.pkl'), env)
    else:
        raise FileNotFoundError(f"vecnormalize.pkl not found in {exp_folder}")


def parse_noise(hyperparams, n_actions, algo):
    noise_type = hyperparams["noise_type"].strip()
    noise_std = hyperparams["noise_std"]
    if "adaptive-param" in noise_type:
        assert algo == "ddpg", "Parameter is not supported by SAC"
        hyperparams["param_noise"] = AdaptiveParamNoiseSpec(
            initial_stddev=noise_std, desired_action_stddev=noise_std
        )
    elif "normal" in noise_type:
        if "lin" in noise_type:
            hyperparams["action_noise"] = LinearNormalActionNoise(
                mean=np.zeros(n_actions),
                sigma=noise_std * np.ones(n_actions),
                final_sigma=hyperparams.get("noise_std_final", 0.0) * np.ones(n_actions),
                max_steps=args.n_timesteps,
            )
        else:
            hyperparams["action_noise"] = NormalActionNoise(
                mean=np.zeros(n_actions), sigma=noise_std * np.ones(n_actions)
            )
    elif "ornstein-uhlenbeck" in noise_type:
        hyperparams["action_noise"] = OrnsteinUhlenbeckActionNoise(
            mean=np.zeros(n_actions), sigma=noise_std * np.ones(n_actions)
        )
    else:
        raise RuntimeError(f"Unknown noise type {noise_type}")

    del hyperparams["noise_type"]
    del hyperparams["noise_std"]
    hyperparams = del_key_from_hyperparams(hyperparams, "noise_std_final")
    LOGGER.info(f"Applying {noise_type} noise with std {noise_std}")

    return hyperparams


def continue_learning(hyperparams, args, env, tensorboard_log, callbacks, normalize, save_path, env_id, rank):

    # Policy should not be changed
    del hyperparams["policy"]
    hyperparams = del_key_from_hyperparams(hyperparams, "n_envs")
    exp_folder = os.path.dirname(args.trained_agent)

    # TODO: The following statements have no effect, probably the model definition should be moved after these lines
    if normalize:
        env = create_vec_normalized_env(exp_folder, env)

    LOGGER.info(f"Loading from {args.trained_agent}")
    model = ALGOS[args.algo].load(args.trained_agent, env=env, tensorboard_log=tensorboard_log,
                                  verbose=args.logging_mode.value, **hyperparams)

    # Arguments to the learn function
    kwargs = {}
    if args.log_interval > -1:
        kwargs = {"log_interval": args.log_interval}
    if len(callbacks) > 0:
        kwargs["callback"] = callbacks

    try:
        model.learn(args.n_timesteps, **kwargs)
    except KeyboardInterrupt:
        pass

    # Only save worker of rank 0 when using mpi
    if rank == 0:
        LOGGER.info(f"Saving model to {save_path}")
        model.save(os.path.join(save_path, str(env_id)))
        if normalize:
            # Important: save the running average, for testing the agent we need that normalization
            LOGGER.info(f"Saving vectorized and normalized environment wrapper to {save_path}")
            model.get_vec_normalize_env().save(os.path.join(save_path, "vecnormalize.pkl"))


def optimize_parameters(hyperparams, args, save_path, create_env, sampling_setting_reward_configs,
                        env_kwargs, tensorboard_log, sampling_setting_observation_configs):
    def create_model(hyperparams, configs):
        """
        Helper to create a model with different hyperparameters
        """
        hyperparams = del_key_from_hyperparams(hyperparams, "n_envs")

        return ALGOS[args.algo](env=create_env(args.n_envs, eval_env=False, **configs),
                                tensorboard_log=tensorboard_log, verbose=1, **hyperparams)

    if args.optimize_hyperparams:
        LOGGER.debug("Optimizing model hyperparameters")
        log_path = os.path.join(save_path, "model_hyperparameter_optimization")
        os.makedirs(log_path, exist_ok=True)
        optimized_hyperparams = optimize_hyperparams(args.algo,
                                                     args.env,
                                                     create_model,
                                                     create_env,
                                                     n_timesteps=args.n_timesteps,
                                                     eval_freq=args.eval_freq,
                                                     n_eval_episodes=args.eval_episodes,
                                                     n_trials=args.n_trials,
                                                     hyperparams=hyperparams,
                                                     configs=env_kwargs,
                                                     n_jobs=args.n_jobs,
                                                     seed=args.seed,
                                                     sampler_method=args.sampler,
                                                     pruner_method=args.pruner,
                                                     verbose=args.logging_mode.value,
                                                     log_path=log_path,
                                                     best_model_save_path=log_path)
        LOGGER.info(f"Saving optimized model hyperparameters to {log_path}")
        report_name = f"report_{args.algo}_{args.env}-{args.n_trials}-" \
                      f"trials-{args.n_timesteps}-steps-{args.sampler}-{args.pruner}.yml"
        with open(os.path.join(log_path, report_name), "w") as f:
            yaml.dump(optimized_hyperparams, f)

    if args.optimize_reward_configs:
        LOGGER.debug("Optimizing reward configurations")

        log_path = os.path.join(save_path, "reward_configuration_optimization")
        os.makedirs(log_path, exist_ok=True)
        optimized_reward_configs = optimize_reward_configs(args.algo,
                                                           args.env,
                                                           create_model,
                                                           create_env,
                                                           sampling_setting=sampling_setting_reward_configs,
                                                           n_timesteps=args.n_timesteps,
                                                           eval_freq=args.eval_freq,
                                                           n_eval_episodes=args.eval_episodes,
                                                           n_trials=args.n_trials,
                                                           hyperparams=hyperparams,
                                                           configs=env_kwargs,
                                                           n_jobs=args.n_jobs,
                                                           seed=args.seed,
                                                           sampler_method=args.sampler,
                                                           pruner_method=args.pruner,
                                                           verbose=args.logging_mode.value,
                                                           log_path=log_path,
                                                           best_model_save_path=log_path,
                                                           guided=args.guided)

        LOGGER.info(f"Saving optimized reward configurations to {log_path}")
        report_name = f"report_{args.algo}_{args.env}-{args.n_trials}-trials" \
                      f"-{args.n_timesteps}-steps-{args.sampler}-{args.pruner}.yml"
        with open(os.path.join(log_path, report_name), "w") as f:
            yaml.dump(optimized_reward_configs, f)

    if args.optimize_observation_configs:
        LOGGER.debug("Optimizing observation configurations")

        log_path = os.path.join(save_path, "observation_configuration_optimization")
        os.makedirs(log_path, exist_ok=True)
        optimized_observation = optimize_observation_configs(args.algo,
                                                             args.env,
                                                             create_model,
                                                             create_env,
                                                             sampling_setting=sampling_setting_observation_configs,
                                                             n_timesteps=args.n_timesteps,
                                                             eval_freq=args.eval_freq,
                                                             n_eval_episodes=args.eval_episodes,
                                                             n_trials=args.n_trials,
                                                             hyperparams=hyperparams,
                                                             configs=env_kwargs,
                                                             n_jobs=args.n_jobs,
                                                             seed=args.seed,
                                                             sampler_method=args.sampler,
                                                             pruner_method=args.pruner,
                                                             verbose=args.logging_mode.value,
                                                             log_path=log_path,
                                                             best_model_save_path=log_path)

        LOGGER.info(f"Saving optimized observation configurations to {log_path}")
        report_name = f"report_{args.algo}_{args.env}-{args.n_trials}-trials" \
                      f"-{args.n_timesteps}-steps-{args.sampler}-{args.pruner}.yml"
        with open(os.path.join(log_path, report_name), "w") as f:
            yaml.dump(optimized_observation, f)


def train_from_scratch(hyperparams, args, env, tensorboard_log, callbacks, save_path, env_id, normalize, rank):
    hyperparams = del_key_from_hyperparams(hyperparams, "n_envs")
    ### add gail model
    if args.algo == 'gail':
        expert_path = os.path.join(args.save_path, "../expert/R_G3_T3_36.npz")
        dataset = ExpertDataset(expert_path, verbose=1)
        tb_log_name = hyperparams['tb_log_name']
        hyperparams = del_key_from_hyperparams(hyperparams, "mode")
        hyperparams = del_key_from_hyperparams(hyperparams, "rule")
        hyperparams = del_key_from_hyperparams(hyperparams, "tb_log_name")
        model = ALGOS[args.algo](env=env, expert_dataset=dataset, tensorboard_log=tensorboard_log,
                             seed=args.seed, verbose=1, **hyperparams)
    else:
        model = ALGOS[args.algo](env=env, tensorboard_log=tensorboard_log,
                             seed=args.seed, verbose=args.logging_mode.value, **hyperparams)

    # Arguments to the learn function
    kwargs = {}
    if args.log_interval > -1:
        kwargs = {"log_interval": args.log_interval}
    if len(callbacks) > 0:
        kwargs["callback"] = callbacks
    ### add gail model 
    if args.algo == 'gail':
        kwargs['tb_log_name'] = tb_log_name

    try:
        model.learn(args.n_timesteps, **kwargs)
    except KeyboardInterrupt:
        pass

    # Only save worker of rank 0 when using mpi
    if rank == 0:
        LOGGER.info(f"Saving model to {save_path}")
        model.save(os.path.join(save_path, str(env_id)))
        if normalize:
            # Important: save the running average, for testing the agent we need that normalization
            LOGGER.info(f"Saving vectorized and normalized environment wrapper to {save_path}")
            model.get_vec_normalize_env().save(os.path.join(save_path, "vecnormalize.pkl"))


def del_key_from_hyperparams(hyperparams: dict, key: str):
    if key in hyperparams.keys():
        del hyperparams[key]

    return hyperparams


def run_stable_baselines(args):
    """
    Run training with stable baselines
    For help, see README.md
    """
    t1 = time.time()
    args.info_keywords = tuple(args.info_keywords)

    # Check support for algo
    if ALGOS[args.algo] is None:
        raise ValueError(f"{args.algo} requires MPI to be installed")

    # TODO: check subfolder existence for commonroad env

    # Set save paths if not specified already
    save_path = args.save_path if args.save_path is not None else construct_save_path(args)
    os.makedirs(save_path, exist_ok=True)
    construct_logger(args.logging_mode.value, save_path, LOGGER)

    # Configure Stable Baselines logger
    # For details of formats, see https://github.com/hill-a/stable-baselines/blob/
    # b3f414f4f2900403107357a2206f80868af16da3/stable_baselines/logger.py#L572
    logger.configure(folder=save_path)

    if args.env == "cr-monitor-v0":
        import gym_monitor
        import crmonitor

    # Going through custom gym packages to let them register in the global registory
    for env_module in args.gym_packages:
        importlib.import_module(env_module)

    env_id = args.env
    registered_envs = set(gym.envs.registry.env_specs.keys())

    # If the environment is not found, suggest the closest match
    if env_id not in registered_envs:
        try:
            closest_match = difflib.get_close_matches(env_id, registered_envs, n=1)[0]
        except IndexError:
            closest_match = "'no close match found...'"
        raise ValueError(f"{env_id} not found in gym registry, you maybe meant {closest_match}?")

    if args.seed < 0:
        # Seed but with a random one
        args.seed = np.random.randint(2 ** 32 - 1)

    set_global_seeds(args.seed)

    if args.trained_agent != "":
        valid_extension = args.trained_agent.endswith(".pkl") or args.trained_agent.endswith(".zip")
        assert valid_extension and os.path.isfile(args.trained_agent), \
            "The trained_agent must be a valid path to a .zip/.pkl file"

    rank = 0
    if mpi4py is not None and MPI.COMM_WORLD.Get_size() > 1:
        LOGGER.info(f"Using MPI for multiprocessing with {MPI.COMM_WORLD.Get_size()} workers")
        rank = MPI.COMM_WORLD.Get_rank()
        LOGGER.info(f"Worker rank: {rank}")

        args.seed += rank
        if rank != 0:
            # Do not log anything for non-master nodes
            LOGGER.setLevel(logging.WARNING)
            args.tensorboard_log = ""

    tensorboard_log = None if args.tensorboard_log == "" else os.path.join(args.tensorboard_log, env_id)

    is_atari = False
    if "NoFrameskip" in env_id:
        is_atari = True

    LOGGER.info(f"Environment id: {env_id}")
    LOGGER.info(f"Seed: {args.seed}")
    LOGGER.info(f"Using {args.n_envs} environments")
    LOGGER.info(f"Learning with {args.n_timesteps} timesteps")

    env_kwargs = {}
    if "commonroad" in env_id or env_id == "cr-monitor-v0":
        # Get environment keyword arguments including observation and reward configurations
        with open(PATH_PARAMS["configs"][env_id], "r") as config_file:
            configs = yaml.safe_load(config_file)
            env_configs = configs["env_configs"]
            sampling_setting_reward_configs = configs["sampling_setting_reward_configs"]
            sampling_setting_observation_configs = configs["sampling_setting_observation_configs"]
            env_kwargs.update(env_configs)

    # Overwrite environment configurations if needed, first from file then from command arguments -> TODO: remove???
    if os.path.isfile(args.configs_path):
        with open(args.configs_path, "r") as configs_file:
            env_kwargs.update(yaml.safe_load(configs_file))

    if args.env_kwargs is not None:
        env_kwargs.update(args.env_kwargs)

    # Save environment configurations for later inspection
    LOGGER.info(f"Saving environment configurations into {save_path}")
    with open(os.path.join(save_path, "environment_configurations.yml"), "w") as output_file:
        yaml.dump(env_kwargs, output_file)
    LOGGER.debug(f"Environment configurations:")
    LOGGER.debug(pformat(env_kwargs))

    # Load model hyperparameters from yaml file
    with open(os.path.join(ROOT_STR, f"commonroad_rl/hyperparams/{args.algo}.yml"), "r") as f:
        hyperparams_dict = yaml.safe_load(f)
        if env_id in list(hyperparams_dict.keys()):
            hyperparams = hyperparams_dict[env_id]
        elif is_atari:
            hyperparams = hyperparams_dict["atari"]
        else:
            raise ValueError(f"Model hyperparameters not found for {args.algo}-{env_id}")

    # Overwrite model hyperparameters if needed, first from file then from command arguments
    if os.path.isfile(args.hyperparams_path):
        with open(args.hyperparams_path, "r") as hyperparams_file:
            tmp = yaml.safe_load(hyperparams_file)
            hyperparams.update(tmp)
    if args.hyperparams is not None:
        hyperparams.update(args.hyperparams)

    # Save model hyperparameters for later inspection
    LOGGER.info(f"Saving model hyperparameters into {save_path}")
    with open(os.path.join(save_path, "model_hyperparameters.yml"), "w") as f:
        yaml.dump(hyperparams, f)
    LOGGER.debug(f"Model hyperparameters loaded and set for {args.algo}-{env_id}")
    LOGGER.debug(pformat(hyperparams))

    # HER is only a wrapper around an algo
    algo = args.algo
    if algo == "her":
        algo = hyperparams["model_class"]
        assert algo in {"sac", "ddpg", "dqn", "td3"}, f"{algo} is not compatible with HER"
        # Retrieve the model class
        hyperparams["model_class"] = ALGOS[hyperparams["model_class"]]
        if hyperparams["model_class"] is None:
            raise ValueError(f"{algo} requires MPI to be installed")

    # Create learning rate schedules for ppo2 and sac
    if algo in ["ppo2", "sac", "td3"]:
        hyperparams = create_parameter_schedule(hyperparams)

    # Check if normalize
    normalize = False
    normalize_kwargs = {}
    if "normalize" in hyperparams.keys():
        normalize = hyperparams["normalize"]
        if isinstance(normalize, str):
            normalize_kwargs = eval(normalize)
            normalize = True
        del hyperparams["normalize"]

    # Convert to python object if needed
    if "policy_kwargs" in hyperparams.keys() and isinstance(hyperparams["policy_kwargs"], str):
        hyperparams["policy_kwargs"] = eval(hyperparams["policy_kwargs"])

    # Delete keys so the dict can be pass to the model constructor
    hyperparams = del_key_from_hyperparams(hyperparams, "n_timesteps")

    # Obtain a class object from a wrapper name string in hyperparams and delete the entry
    env_wrapper = get_wrapper_class(hyperparams)
    hyperparams = del_key_from_hyperparams(hyperparams, "env_wrapper")

    callbacks = []
    if args.save_freq > 0:
        # Account for the number of parallel environments
        args.save_freq = max(args.save_freq // args.n_envs, 1)
        callbacks.append(CheckpointCallback(save_freq=args.save_freq, save_path=save_path,
                                            name_prefix="rl_model", verbose=args.logging_mode.value))
        callbacks.append(SaveVecNormalizeCallback(
            save_freq=args.save_freq,
            save_path=save_path,
            name_prefix="vecnormalize",
            verbose=args.logging_mode.value,
        ))

    def create_env(n_envs, eval_env=False, **kwargs):
        """
        Create the environment and wrap it if necessary

        :param n_envs: (int)
        :param eval_env: (bool) Whether is it an environment used for evaluation or not
        :return: (Union[gym.Env, VecEnv])
        :return: (gym.Env)
        """

        # Set verbosity
        env_kwargs["logging_mode"] = args.logging_mode.value

        # Update environment keyword arguments from optimization sampler if specified
        env_kwargs.update(kwargs)

        # Differentiate keyword arguments for test environments
        env_kwargs_test = copy.deepcopy(env_kwargs)
        if "commonroad" in args.env or args.env == "cr-monitor-v0":
            env_kwargs_test["test_env"] = True

        # Do not log eval env (issue with writing the same file)
        # log_dir = None if eval_env else save_path
        log_dir = os.path.join(save_path, "test") if eval_env else save_path

        if is_atari:
            LOGGER.debug("Using Atari wrapper")
            new_env = make_atari_env(env_id, num_env=n_envs, seed=args.seed)
            # Frame-stacking with 4 frames
            new_env = VecFrameStack(new_env, n_stack=4)
        elif algo in ["dqn", "ddpg"]:
            if hyperparams.get("normalize", False):
                LOGGER.warning("WARNING: normalization not supported yet for DDPG/DQN")

            new_env = DummyVecEnv([make_env(env_id,
                                            rank,
                                            args.seed,
                                            wrapper_class=env_wrapper,
                                            log_dir=log_dir,
                                            logging_path=save_path,
                                            env_kwargs=env_kwargs_test if eval_env else env_kwargs,
                                            info_keywords=args.info_keywords)])
            new_env.seed(args.seed)
            if env_wrapper is not None:
                new_env = env_wrapper(new_env)
        elif algo != 'gail':
            if n_envs == 1:
                new_env = DummyVecEnv([make_env(env_id,
                                                0,
                                                args.seed,
                                                wrapper_class=env_wrapper,
                                                log_dir=log_dir,
                                                logging_path=save_path,
                                                env_kwargs=env_kwargs_test if eval_env else env_kwargs,
                                                info_keywords=args.info_keywords)])
            else:
                # env = SubprocVecEnv([make_env(env_id, i, args.seed) for i in range(n_envs)])
                # On most env, SubprocVecEnv does not help and is quite memory hungry
                new_env = SubprocVecEnv([make_env(env_id,
                                                  i,
                                                  args.seed + n_envs if eval_env else args.seed,
                                                  log_dir=log_dir,
                                                  logging_path=save_path,
                                                  wrapper_class=env_wrapper,
                                                  env_kwargs=env_kwargs_test if eval_env else env_kwargs,
                                                  subproc=True,
                                                  info_keywords=args.info_keywords) for i in range(n_envs)],
                                        start_method="spawn")
            if normalize:
                if len(normalize_kwargs) > 0:
                    LOGGER.debug(f"Normalization activated: {normalize_kwargs}")
                else:
                    LOGGER.debug("Normalizing observations and rewards")
                new_env = VecNormalize(new_env, **normalize_kwargs)
        # Optional Frame-stacking
        if hyperparams.get("frame_stack", False):
            n_stack = hyperparams["frame_stack"]
            new_env = VecFrameStack(new_env, n_stack)
            LOGGER.info(f"Stacking {n_stack} frames")
            del hyperparams["frame_stack"]
            output_file.close()
        if args.algo == "her":
            # Wrap the env if need to flatten the dict obs
            if isinstance(new_env, VecEnv):
                new_env = _UnvecWrapper(new_env)
            new_env = HERGoalEnvWrapper(new_env)
        ### add gail model
        if args.algo == "gail":
            env1 = gym.make("commonroad-v1",
                            **env_configs)
            print('env1.observation_space', env1.observation_space)        
                     
            env_kwargs["mode"] = hyperparams["mode"]
            env_kwargs["rule"] = hyperparams["rule"]
            env_kwargs["env"] = env1
            new_env = DummyVecEnv([make_env(env_id,
                                        0,
                                        args.seed,
                                        log_dir=log_dir,
                                        logging_path=save_path,
                                        env_kwargs=env_kwargs_test if eval_env else env_kwargs,
                                        info_keywords=args.info_keywords)])
        return new_env

    # HINT: Three main options to be chosen from
    # HINT: 1. Continue training with a pretrained agent
    # HINT: 2. Optimize model hyperparameters and/or configurations of observations and rewards
    # HINT: 3. Start training from scratch
    if args.optimize_hyperparams or args.optimize_observation_configs or args.optimize_reward_configs:
        optimize_parameters(hyperparams, args, save_path, create_env, sampling_setting_reward_configs,
                            env_kwargs, tensorboard_log, sampling_setting_observation_configs)
    else:
        # Create training environments
        env = create_env(args.n_envs)

        # Create testing environments if needed, do not normalize reward
        if args.eval_freq > 0:
            # Account for the number of parallel environments
            eval_freq = max(args.eval_freq // args.n_envs, 1)

            # Do not normalize the rewards of the eval env
            old_kwargs = None
            if normalize:
                if len(normalize_kwargs) > 0:
                    old_kwargs = normalize_kwargs.copy()
                    normalize_kwargs["norm_reward"] = False
                else:
                    normalize_kwargs = {"norm_reward": False}

            save_vec_normalize = SaveVecNormalizeCallback(
                save_freq=1,
                save_path=save_path,
                name_prefix="vecnormalize",
                verbose=args.logging_mode.value,
            )

            eval_callback = MultiEnvsEvalCallback(
                create_env(args.n_envs, eval_env=True),
                log_path=save_path,
                best_model_save_path=save_path,
                eval_freq=eval_freq,
                n_eval_timesteps=args.eval_timesteps,
                callback_on_new_best=save_vec_normalize,
                verbose=args.logging_mode.value,
            )
            callbacks.append(eval_callback)

            # Restore original kwargs
            if old_kwargs is not None:
                normalize_kwargs = old_kwargs.copy()

        LOGGER.info(f"Elapsed time for preparing steps: {time.time() - t1} s")

        # Parse noise string for DDPG and SAC and TD3
        if algo in ["ddpg", "sac", "td3"] and hyperparams.get("noise_type") is not None:
            hyperparams = parse_noise(hyperparams, env.action_space.shape[0], algo)

        # Load a trained model and continue training
        if os.path.isfile(args.trained_agent):
            continue_learning(hyperparams, args, env, tensorboard_log, callbacks, normalize, save_path, env_id, rank)
        else:
            # Train an agent from scratch
            train_from_scratch(hyperparams, args, env, tensorboard_log, callbacks, save_path, env_id, normalize, rank)

    LOGGER.info(f"Elapsed time: {time.time() - t1} s")


if __name__ == "__main__":
    args = run_stable_baselines_argsparser().parse_args(sys.argv[1:])
    run_stable_baselines(args)



#%matplotlib inline

import os
import argparse
import matplotlib.pyplot as plt
from IPython import display

# import functions to read xml file and visualize commonroad objects
from commonroad.common.file_reader import CommonRoadFileReader
from commonroad.visualization.mp_renderer import MPRenderer

def argsparser():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("path", help="path to xml file", type=str)
    parser.add_argument("--num_time_steps", "-n", type=int,  default=40,
                        help="number of timesteps to plot")
    parser.add_argument("--save_location", "-s", type=str, default=None, help="Path to directory to save rendered images to")
    parser.add_argument("--time_granularity", "-tg", type=int, default=1, help="Granularity for amount of timesteps in between plots")
    parser.add_argument("--start_at_time_step", "-st", type=int, default=0, help="Timestep to start plotting from (default is 0)")
    return parser.parse_args()

def main():
    args = argsparser()

    # generate path of the file to be opened
    file_path = args.path
    print(file_path)

    # read in the scenario and planning problem set
    scenario, planning_problem_set = CommonRoadFileReader(file_path).open()

    # plot the scenario for 40 time step, here each time step corresponds to 0.1 second
    start = args.start_at_time_step
    length = args.num_time_steps
    gran = args.time_granularity

    for i in range(start, length):
        plt.figure(figsize=(25, 10))
        rnd = MPRenderer()
        # plot the scenario at different time step
        scenario.draw(rnd, draw_params={'time_begin': gran*i + start})
        # plot the planning problem set
        planning_problem_set.draw(rnd)
        
        rnd.render()
        if not (args.save_location is None):
            os.makedirs(args.save_location, exist_ok = True)
            plt.savefig(os.path.join(args.save_location,f"{scenario.scenario_id}-{i}.png"))
    plt.show()

if __name__ == "__main__":
    main()


"""
A utility function to be called when optimizing model hyperparameters
"""
import logging
import os
from pprint import pformat

import numpy as np
import optuna
import yaml
from optuna.integration.skopt import SkoptSampler
from optuna.pruners import SuccessiveHalvingPruner, MedianPruner
from optuna.samplers import RandomSampler, TPESampler
from stable_baselines.common.base_class import _UnvecWrapper
from stable_baselines.common.vec_env import VecEnv
from stable_baselines.ddpg import (
    AdaptiveParamNoiseSpec,
    NormalActionNoise,
    OrnsteinUhlenbeckActionNoise,
)
from stable_baselines.her import HERGoalEnvWrapper

from commonroad_rl.utils_run.callbacks import HyperparamsTrialEvalCallback

LOGGER = logging.getLogger(__name__)


def optimize_hyperparams(
    algo,
    env,
    model_fn,
    env_fn,
    n_timesteps=5000,
    eval_freq=500,
    n_eval_episodes=5,
    n_trials=10,
    hyperparams=None,
    configs=None,
    n_jobs=1,
    sampler_method="random",
    pruner_method="halving",
    seed=13,
    verbose=1,
    log_path=None,
    best_model_save_path=None,
):
    """

    :param algo: (str)
    :param env: (str)
    :param model_fn: (func) function that is used to instantiate the model
    :param env_fn: (func) function that is used to instantiate the env
    :param n_trials: (int) maximum number of trials for finding the best hyperparams
    :param n_timesteps: (int) maximum number of timesteps per trial
    :param hyperparams: (dict) model hyperparameters to be optimized
    :param configs: (dict) environment configurations
    :param n_jobs: (int) number of parallel jobs
    :param sampler_method: (str)
    :param pruner_method: (str)
    :param seed: (int)
    :param verbose: (int)
    :param log_path: (str) folder for saving evaluation results during optimization
    :param best_model_save_path (str) folder for saving the best model
    :return: (dict) detailed result of the optimization
    """
    # TODO: eval each hyperparams several times to account for noisy evaluation
    # TODO: take into account the normalization (also for the test env -> sync obs_rms)
    # Set logger if not yet exists
    if verbose == 2:
        LOGGER.setLevel(logging.DEBUG)
    elif verbose == 1:
        LOGGER.setLevel(logging.INFO)
    else:
        LOGGER.setLevel(logging.ERROR)

    if not len(LOGGER.handlers):
        formatter = logging.Formatter("[%(levelname)s] %(name)s - %(message)s")
        stream_handler = logging.StreamHandler()
        if verbose == 2:
            stream_handler.setLevel(logging.DEBUG)
        elif verbose == 1:
            stream_handler.setLevel(logging.INFO)
        else:
            stream_handler.setLevel(logging.ERROR)
        stream_handler.setFormatter(formatter)
        LOGGER.addHandler(stream_handler)

        if log_path is not None:
            file_handler = logging.FileHandler(
                filename=os.path.join(log_path, "console_copy.txt")
            )
            if verbose == 2:
                file_handler.setLevel(logging.DEBUG)
            elif verbose == 1:
                file_handler.setLevel(logging.INFO)
            else:
                file_handler.setLevel(logging.ERROR)
            file_handler.setFormatter(formatter)
            LOGGER.addHandler(file_handler)

    LOGGER.info("Optimizing model hyperparameters for {}".format(algo))

    # Model hyperparameters such as learning rate, minibatch size, ... etc
    if hyperparams is None:
        hyperparams = {}

    n_startup_trials = 10
    n_evaluations = n_timesteps // eval_freq

    LOGGER.info(
        "Optimizing with {} trials using {} parallel jobs, each with {} maximal time steps and {} evaluations".format(
            n_trials, n_jobs, n_timesteps, n_evaluations
        )
    )

    # n_warmup_steps: Disable pruner until the trial reaches the given number of step.
    if sampler_method == "random":
        sampler = RandomSampler(seed=seed)
    elif sampler_method == "tpe":
        sampler = TPESampler(n_startup_trials=n_startup_trials, seed=seed)
    elif sampler_method == "skopt":
        # cf https://scikit-optimize.github.io/#skopt.Optimizer
        # GP: gaussian process
        # Gradient boosted regression: GBRT
        sampler = SkoptSampler(
            skopt_kwargs={"base_estimator": "GP", "acq_func": "gp_hedge"}
        )
    else:
        raise ValueError(
            "[hyperparams_opt.py] Unknown sampler: {}".format(sampler_method)
        )

    if pruner_method == "halving":
        pruner = SuccessiveHalvingPruner(
            min_resource=1, reduction_factor=4, min_early_stopping_rate=0
        )
    elif pruner_method == "median":
        pruner = MedianPruner(
            n_startup_trials=n_startup_trials, n_warmup_steps=n_evaluations // 3
        )
    elif pruner_method == "none":
        # Do not prune
        pruner = MedianPruner(n_startup_trials=n_trials, n_warmup_steps=n_evaluations)
    else:
        raise ValueError(
            "[hyperparams_opt.py] Unknown pruner: {}".format(pruner_method)
        )

    LOGGER.debug("Sampler: {} - Pruner: {}".format(sampler_method, pruner_method))

    # Create study object on model hyperparameters from Optuna
    hyperparams_study = optuna.create_study(sampler=sampler, pruner=pruner)

    # Prepare the sampler
    algo_sampler = HYPERPARAMS_SAMPLER[algo]

    def objective_hyperparams(trial):
        """
        Optimization objective for model hyperparameters
        """
        trial.model_class = None

        if algo == "her":
            trial.model_class = hyperparams["model_class"]

        # Hack to use DDPG/TD3 noise sampler
        if algo in ["ddpg", "td3"] or trial.model_class in ["ddpg", "td3"]:
            trial.n_actions = env_fn(n_envs=1).action_space.shape[0]

        # Get keyword arguments for model hyperparameters and environment configurations
        kwargs_hyperparams = hyperparams.copy()
        kwargs_configs = configs.copy()

        # Sample model hyperparameters and keep environment configurations
        sampled_hyperparams = algo_sampler(trial)
        kwargs_hyperparams.update(sampled_hyperparams)

        # Save data for later inspection
        tmp_path = os.path.join(log_path, "trial_" + str(trial.number))
        os.makedirs(tmp_path, exist_ok=True)
        with open(
            os.path.join(tmp_path, "sampled_model_hyperparameters.yml"), "w"
        ) as f:
            LOGGER.info(
                "Saving sampled observation configurations into {}".format(tmp_path)
            )
            yaml.dump(kwargs_hyperparams, f)
        LOGGER.debug("Sampled model hyperparameters:")
        LOGGER.debug(pformat(kwargs_hyperparams))

        # Create model and environments for optimization
        model = model_fn(kwargs_hyperparams, kwargs_configs)
        eval_env = env_fn(n_envs=1, eval_env=True, **kwargs_configs)

        # Account for parallel envs
        eval_freq_ = eval_freq
        if isinstance(model.get_env(), VecEnv):
            eval_freq_ = max(eval_freq // model.get_env().num_envs, 1)

        LOGGER.info(
            "Evaluating with {} episodes after every {} time steps".format(
                n_eval_episodes, eval_freq_
            )
        )

        # TODO: use non-deterministic eval for Atari?
        hyperparams_eval_callback = HyperparamsTrialEvalCallback(
            eval_env,
            trial,
            n_eval_episodes=n_eval_episodes,
            eval_freq=eval_freq_,
            log_path=log_path,
            best_model_save_path=best_model_save_path,
            deterministic=True,
            verbose=verbose,
        )
        if algo == "her":
            # Wrap the env if need to flatten the dict obs
            if isinstance(eval_env, VecEnv):
                eval_env = _UnvecWrapper(eval_env)
            eval_env = HERGoalEnvWrapper(eval_env)

        try:
            model.learn(n_timesteps, callback=hyperparams_eval_callback)
            # Free memory
            model.env.close()
            eval_env.close()
        except AssertionError:
            # Sometimes, random hyperparams can generate NaN
            # Free memory
            model.env.close()
            eval_env.close()
            raise optuna.exceptions.TrialPruned()

        is_pruned = hyperparams_eval_callback.is_pruned
        cost = hyperparams_eval_callback.cost
        del model.env, eval_env
        del model
        if is_pruned:
            raise optuna.exceptions.TrialPruned()
        return cost

    try:
        hyperparams_study.optimize(
            objective_hyperparams, n_trials=n_trials, n_jobs=n_jobs
        )
    except KeyboardInterrupt:
        pass

    LOGGER.info("Number of finished trials: {}".format(len(hyperparams_study.trials)))
    LOGGER.info("Best value: {}".format(hyperparams_study.best_trial.value))
    LOGGER.info("Best model hyperparameters: ")
    LOGGER.info(pformat(hyperparams_study.best_trial.params))

    return hyperparams_study.best_trial.params


def sample_ppo2_params(trial):
    """
    Sampler for PPO2 hyperparams.

    :param trial: (optuna.trial)
    :return: (dict)
    """
    batch_size = trial.suggest_categorical("batch_size", [32, 64, 128, 256])
    n_steps = trial.suggest_categorical(
        "n_steps", [16, 32, 64, 128, 256, 512, 1024, 2048]
    )
    gamma = trial.suggest_categorical(
        "gamma", [0.9, 0.95, 0.98, 0.99, 0.995, 0.999, 0.9999]
    )
    learning_rate = trial.suggest_loguniform("lr", 1e-5, 1)
    ent_coef = trial.suggest_loguniform("ent_coef", 0.00000001, 0.1)
    cliprange = trial.suggest_categorical("cliprange", [0.1, 0.2, 0.3, 0.4])
    noptepochs = trial.suggest_categorical("noptepochs", [1, 5, 10, 20, 30, 50])
    lam = trial.suggest_categorical("lamdba", [0.8, 0.9, 0.92, 0.95, 0.98, 0.99, 1.0])

    if n_steps < batch_size:
        nminibatches = 1
    else:
        nminibatches = int(n_steps / batch_size)

    return {
        "n_steps": n_steps,
        "nminibatches": nminibatches,
        "gamma": gamma,
        "learning_rate": learning_rate,
        "ent_coef": ent_coef,
        "cliprange": cliprange,
        "noptepochs": noptepochs,
        "lam": lam,
    }


def sample_a2c_params(trial):
    """
    Sampler for A2C hyperparams.

    :param trial: (optuna.trial)
    :return: (dict)
    """
    gamma = trial.suggest_categorical(
        "gamma", [0.9, 0.95, 0.98, 0.99, 0.995, 0.999, 0.9999]
    )
    n_steps = trial.suggest_categorical(
        "n_steps", [8, 16, 32, 64, 128, 256, 512, 1024, 2048]
    )
    lr_schedule = trial.suggest_categorical("lr_schedule", ["linear", "constant"])
    learning_rate = trial.suggest_loguniform("lr", 1e-5, 1)
    ent_coef = trial.suggest_loguniform("ent_coef", 0.00000001, 0.1)
    vf_coef = trial.suggest_uniform("vf_coef", 0, 1)
    # normalize = trial.suggest_categorical('normalize', [True, False])
    # TODO: take into account the normalization (also for the test env)

    return {
        "n_steps": n_steps,
        "gamma": gamma,
        "learning_rate": learning_rate,
        "lr_schedule": lr_schedule,
        "ent_coef": ent_coef,
        "vf_coef": vf_coef,
    }


def sample_acktr_params(trial):
    """
    Sampler for ACKTR hyperparams.

    :param trial: (optuna.trial)
    :return: (dict)
    """
    gamma = trial.suggest_categorical(
        "gamma", [0.9, 0.95, 0.98, 0.99, 0.995, 0.999, 0.9999]
    )
    n_steps = trial.suggest_categorical(
        "n_steps", [16, 32, 64, 128, 256, 512, 1024, 2048]
    )
    lr_schedule = trial.suggest_categorical("lr_schedule", ["linear", "constant"])
    learning_rate = trial.suggest_loguniform("lr", 1e-5, 1)
    ent_coef = trial.suggest_loguniform("ent_coef", 0.00000001, 0.1)
    vf_coef = trial.suggest_uniform("vf_coef", 0, 1)

    return {
        "n_steps": n_steps,
        "gamma": gamma,
        "learning_rate": learning_rate,
        "lr_schedule": lr_schedule,
        "ent_coef": ent_coef,
        "vf_coef": vf_coef,
    }


def sample_sac_params(trial):
    """
    Sampler for SAC hyperparams.

    :param trial: (optuna.trial)
    :return: (dict)
    """
    gamma = trial.suggest_categorical(
        "gamma", [0.9, 0.95, 0.98, 0.99, 0.995, 0.999, 0.9999]
    )
    learning_rate = trial.suggest_loguniform("lr", 1e-5, 1)
    batch_size = trial.suggest_categorical("batch_size", [16, 32, 64, 128, 256, 512])
    buffer_size = trial.suggest_categorical(
        "buffer_size", [int(1e4), int(1e5), int(1e6)]
    )
    learning_starts = trial.suggest_categorical(
        "learning_starts", [0, 1000, 10000, 20000]
    )
    train_freq = trial.suggest_categorical("train_freq", [1, 10, 100, 300])
    # gradient_steps takes too much time
    # gradient_steps = trial.suggest_categorical('gradient_steps', [1, 100, 300])
    gradient_steps = train_freq
    ent_coef = trial.suggest_categorical(
        "ent_coef", ["auto", 0.5, 0.1, 0.05, 0.01, 0.0001]
    )
    net_arch = trial.suggest_categorical("net_arch", ["small", "medium", "big"])

    net_arch = {
        "small": [64, 64],
        "medium": [256, 256],
        "big": [400, 300],
    }[net_arch]

    target_entropy = "auto"
    if ent_coef == "auto":
        target_entropy = trial.suggest_categorical(
            "target_entropy", ["auto", -1, -10, -20, -50, -100]
        )

    return {
        "gamma": gamma,
        "learning_rate": learning_rate,
        "batch_size": batch_size,
        "buffer_size": buffer_size,
        "learning_starts": learning_starts,
        "train_freq": train_freq,
        "gradient_steps": gradient_steps,
        "ent_coef": ent_coef,
        "target_entropy": target_entropy,
        "policy_kwargs": dict(layers=net_arch),
    }


def sample_td3_params(trial):
    """
    Sampler for TD3 hyperparams.

    :param trial: (optuna.trial)
    :return: (dict)
    """
    gamma = trial.suggest_categorical(
        "gamma", [0.9, 0.95, 0.98, 0.99, 0.995, 0.999, 0.9999]
    )
    learning_rate = trial.suggest_loguniform("lr", 1e-5, 1)
    batch_size = trial.suggest_categorical(
        "batch_size", [16, 32, 64, 100, 128, 256, 512]
    )
    buffer_size = trial.suggest_categorical(
        "buffer_size", [int(1e4), int(1e5), int(1e6)]
    )
    train_freq = trial.suggest_categorical("train_freq", [1, 10, 100, 1000, 2000])
    gradient_steps = train_freq
    noise_type = trial.suggest_categorical(
        "noise_type", ["ornstein-uhlenbeck", "normal"]
    )
    noise_std = trial.suggest_uniform("noise_std", 0, 1)

    hyperparams = {
        "gamma": gamma,
        "learning_rate": learning_rate,
        "batch_size": batch_size,
        "buffer_size": buffer_size,
        "train_freq": train_freq,
        "gradient_steps": gradient_steps,
    }

    if noise_type == "normal":
        hyperparams["action_noise"] = NormalActionNoise(
            mean=np.zeros(trial.n_actions), sigma=noise_std * np.ones(trial.n_actions)
        )
    elif noise_type == "ornstein-uhlenbeck":
        hyperparams["action_noise"] = OrnsteinUhlenbeckActionNoise(
            mean=np.zeros(trial.n_actions), sigma=noise_std * np.ones(trial.n_actions)
        )

    return hyperparams


def sample_trpo_params(trial):
    """
    Sampler for TRPO hyperparams.

    :param trial: (optuna.trial)
    :return: (dict)
    """
    gamma = trial.suggest_categorical(
        "gamma", [0.9, 0.95, 0.98, 0.99, 0.995, 0.999, 0.9999]
    )
    timesteps_per_batch = trial.suggest_categorical(
        "timesteps_per_batch", [16, 32, 64, 128, 256, 512, 1024, 2048, 4096]
    )
    max_kl = trial.suggest_loguniform("max_kl", 0.000001, 1)
    ent_coef = trial.suggest_loguniform("ent_coef", 0.00000001, 0.1)
    lam = trial.suggest_categorical("lamdba", [0.8, 0.9, 0.92, 0.95, 0.98, 0.99, 1.0])
    # cg_damping = trial.suggest_loguniform('cg_damping', 1e-5, 1)
    cg_damping = 0.1
    cg_iters = trial.suggest_categorical("cg_iters", [10, 15, 20, 30])
    vf_stepsize = trial.suggest_loguniform("vf_stepsize", 1e-5, 1)
    vf_iters = trial.suggest_categorical("vf_iters", [1, 3, 5, 10, 20])

    return {
        "gamma": gamma,
        "timesteps_per_batch": timesteps_per_batch,
        "max_kl": max_kl,
        "entcoeff": ent_coef,
        "lam": lam,
        "cg_damping": cg_damping,
        "cg_iters": cg_iters,
        "vf_stepsize": vf_stepsize,
        "vf_iters": vf_iters,
    }


def sample_ddpg_params(trial):
    """
    Sampler for DDPG hyperparams.

    :param trial: (optuna.trial)
    :return: (dict)
    """
    gamma = trial.suggest_categorical(
        "gamma", [0.9, 0.95, 0.98, 0.99, 0.995, 0.999, 0.9999]
    )
    # actor_lr = trial.suggest_loguniform('actor_lr', 1e-5, 1)
    # critic_lr = trial.suggest_loguniform('critic_lr', 1e-5, 1)
    learning_rate = trial.suggest_loguniform("lr", 1e-5, 1)
    batch_size = trial.suggest_categorical("batch_size", [16, 32, 64, 128, 256])
    buffer_size = trial.suggest_categorical(
        "memory_limit", [int(1e4), int(1e5), int(1e6)]
    )
    noise_type = trial.suggest_categorical(
        "noise_type", ["ornstein-uhlenbeck", "normal", "adaptive-param"]
    )
    noise_std = trial.suggest_uniform("noise_std", 0, 1)
    normalize_observations = trial.suggest_categorical(
        "normalize_observations", [True, False]
    )
    normalize_returns = trial.suggest_categorical("normalize_returns", [True, False])

    hyperparams = {
        "gamma": gamma,
        "actor_lr": learning_rate,
        "critic_lr": learning_rate,
        "batch_size": batch_size,
        "memory_limit": buffer_size,
        "normalize_observations": normalize_observations,
        "normalize_returns": normalize_returns,
    }

    if noise_type == "adaptive-param":
        hyperparams["param_noise"] = AdaptiveParamNoiseSpec(
            initial_stddev=noise_std, desired_action_stddev=noise_std
        )
        # Apply layer normalization when using parameter perturbation
        hyperparams["policy_kwargs"] = dict(layer_norm=True)
    elif noise_type == "normal":
        hyperparams["action_noise"] = NormalActionNoise(
            mean=np.zeros(trial.n_actions), sigma=noise_std * np.ones(trial.n_actions)
        )
    elif noise_type == "ornstein-uhlenbeck":
        hyperparams["action_noise"] = OrnsteinUhlenbeckActionNoise(
            mean=np.zeros(trial.n_actions), sigma=noise_std * np.ones(trial.n_actions)
        )
    return hyperparams


def sample_her_params(trial):
    """
    Sampler for HER hyperparams.

    :param trial: (optuna.trial)
    :return: (dict)
    """
    if trial.model_class == SAC:
        hyperparams = sample_sac_params(trial)
    elif trial.model_class == DDPG:
        hyperparams = sample_ddpg_params(trial)
    elif trial.model_class == TD3:
        hyperparams = sample_td3_params(trial)
    else:
        raise NotImplementedError(
            f"Trial model class {trial.model_class} not supported"
        )

    hyperparams["random_exploration"] = trial.suggest_uniform(
        "random_exploration", 0, 1
    )
    hyperparams["n_sampled_goal"] = trial.suggest_categorical(
        "n_sampled_goal", [1, 2, 4, 6, 8]
    )

    return hyperparams


HYPERPARAMS_SAMPLER = {
    "ppo2": sample_ppo2_params,
    "sac": sample_sac_params,
    "a2c": sample_a2c_params,
    "trpo": sample_trpo_params,
    "ddpg": sample_ddpg_params,
    "her": sample_her_params,
    "acktr": sample_acktr_params,
    "td3": sample_td3_params,
}



"""
Module for reading and writing scenario related objects
"""

import os
import re
import random
from pathlib import Path
from typing import List, Tuple

from commonroad.common.file_reader import CommonRoadFileReader
from commonroad.planning.planning_problem import PlanningProblemSet, PlanningProblem
from commonroad.scenario.obstacle import DynamicObstacle
from commonroad.scenario.scenario import Scenario, ScenarioID


def get_project_root():
    """
    Get the root path of project.
    :return: String of the project root path
    """
    return Path(__file__).parent.parent.parent.parent


def get_all_scenario(scenario_path: str) -> List[str]:
    """
    Get the list of all scenarios_backup in the path directory.

    :param scenario_path: the string of the directory
    :return: list of scenarios_backup names
    """
    scenario_list = []
    for file in os.listdir(scenario_path):
        if file.endswith(".xml"):
            scenario_list.append(str(os.path.basename(file)).split(".")[0])
    return scenario_list


def read_scenario(scenario_path: str, scenario_name: str = None) -> Tuple[Scenario, PlanningProblemSet]:
    """
    Get the desired scenario and planning problem set. If the scenario name is not given, return a random scenario with
    its problem set.

    :param scenario_path: path of directory
    :param scenario_name: given scenario name
    :return: scenario and planning problem set
    """
    if scenario_name is None:
        scenario_name = random.choice(get_all_scenario(scenario_path))
    else:
        if not scenario_name.endswith(".xml"):
            scenario_name += ".xml"
    scenario_full_name = os.path.join(scenario_path, scenario_name)
    scenario, planning_problem_set = CommonRoadFileReader(scenario_full_name).open()
    return scenario, planning_problem_set


def read_planning_problem(planning_problem_set: PlanningProblemSet, problem_id: int = None) -> PlanningProblem:
    """
    Get the desired planning problem. If the id is not given, get a random problem.

    :param planning_problem_set: Given planning problem set.
    :param problem_id: The id of the problem
    :return: the planning problem
    """
    if problem_id is None:
        all_ids = list(planning_problem_set.planning_problem_dict.keys())
        planning_problem = planning_problem_set.planning_problem_dict[random.choice(all_ids)]
    else:
        planning_problem = planning_problem_set.planning_problem_dict[problem_id]
    return planning_problem


def scenario_id_to_location(benchmark_id: str) -> str:
    location = re.split(r"([-_])", benchmark_id)[2]
    return location


def restore_scenario(meta_scenario: Scenario, obstacle_list: List[DynamicObstacle], scenario_id: ScenarioID):
    # scenario_new = copy.deepcopy(meta_scenario)
    # scenario_new.add_objects(obstacle_list)
    # return scenario_new
    # TODO: try remove and re-add obstacles ( 2times faster than using deepcopy, 1329ms vs 3514ms)
    meta_scenario.scenario_id = scenario_id
    meta_scenario.remove_obstacle(meta_scenario.obstacles)
    meta_scenario.add_objects(obstacle_list)
    return meta_scenario



import numpy as np
from commonroad.common.solution import VehicleModel
from commonroad.scenario.lanelet import Lanelet
from commonroad.scenario.trajectory import CustomState

from commonroad_rl.gym_commonroad.action.action import ContinuousAction
from commonroad_rl.gym_commonroad.observation import EgoObservation
from commonroad_rl.tests.common.marker import *


def construct_ego_observation(vehicle_model):
    # Ego-related observation settings
    return EgoObservation(
        configs={"ego_configs":
                     {"observe_v_ego": True,
                      "observe_a_ego": True,
                      "observe_relative_heading": True,
                      "observe_steering_angle": True,
                      "observe_global_turn_rate": True,
                      "observe_remaining_steps": True,
                      "observe_is_friction_violation": True},
                 "vehicle_params": {
                     "vehicle_type": 2,
                     "vehicle_model": vehicle_model
                 }})


def prepare_for_ego_test():
    # Initial state of KS vehicle model
    dummy_state_ks = {
        "position": np.array([0.0, 0.0]),
        "yaw_rate": 0.0,
        "slip_angle": 0.0,
        "time_step": 0,
        "orientation": 0.0,
    }

    # Initial state of PM vehicle model
    dummy_state_pm = {
        "position": np.array([0.0, 0.0]),
        "time_step": 0,
    }

    # lanelet config
    lanelet = Lanelet(lanelet_id=0,
                      left_vertices=np.array([[0.0, 3.0], [10.0, 3.0]]),
                      center_vertices=np.array([[0.0, 0.0], [10.0, 0.0]]),
                      right_vertices=np.array([[0.0, -3.0], [10.0, -3.0]]))

    return dummy_state_pm, dummy_state_ks, lanelet


@pytest.mark.parametrize(
    ("ego_angle", "expected_output"),
    [
        (0, 0),
        (1, 1),
        (6, 6 - 2 * np.pi),
        (2 * np.pi, 0),
        (7, 7 - 2 * np.pi),
        (-1, -1),
        (-2 * np.pi, 0),
    ],
)
@unit_test
@functional
def test_get_lane_relative_heading(ego_angle, expected_output):
    dummy_state = {
        "velocity": 0.0,
        "position": np.array([0.0, 0.0]),
        "yaw_rate": 0.0,
        "slip_angle": 0.0,
        "time_step": 0.0,
    }
    ego_vehicle_state = CustomState(**dummy_state, orientation=ego_angle)
    ego_vehicle_lanelet = Lanelet(
        np.array([[0.0, 1.0], [1.0, 1.0]]),
        np.array([[0.0, 0.0], [1.0, 0.0]]),
        np.array([[0.0, -1.0], [1.0, -1.0]]),
        0,
    )
    relative_angle = EgoObservation.get_lane_relative_heading(ego_vehicle_state, ego_vehicle_lanelet)
    assert np.isclose(relative_angle, expected_output)


dummy_state_pm, dummy_state_ks, ego_lanelet = prepare_for_ego_test()


@pytest.mark.parametrize(
    ("velocity", "steering_angle", "vehicle_model", "expected_output"),
    [
        (0, 0, 2, np.array([0., 0., 0., 0., 10.])),
        (10, 0.5, 2, np.array([10., 0., 0.5, 2.118, 10.])),
        (30, 0.5, 2, np.array([30., 0., 0.5, 6.355, 10.])),
        ([0, 0], None, 0, np.array([0., 0., 0., 0., 10.])),
        ([10, 0], None, 0, np.array([10., 0., 0., 0., 10.])),
        ([10, 2], None, 0, np.array([10., 0., 0., 0., 10.])),
    ]
)
@functional
@unit_test
def test_ego_state(velocity, steering_angle, vehicle_model, expected_output):
    vehicle_params = {
        "vehicle_type": 2,  # VehicleType.BMW_320i
        "vehicle_model": vehicle_model,  # 0: VehicleModel.PM; 2: VehicleModel.KS;
    }

    action_configs = {
        "action_type": "continuous", # discrete
        "action_base": "acceleration", # acceleration; jerk
        "long_steps": 5,
        "lat_steps": 5,

    }

    vehicle_action = ContinuousAction(vehicle_params, action_configs)
    if vehicle_params["vehicle_model"] == 2:  # VehicleModel.KS
        initial_state = CustomState(**dummy_state_ks, velocity=velocity, steering_angle=steering_angle, )
    else:  # VehicleModel.PM
        initial_state = CustomState(**dummy_state_pm, velocity=velocity[0], velocity_y=velocity[1])
    vehicle_action.reset(initial_state, dt=1.0)

    # Not to do anything, just continue the way with the given velocity
    action = np.array([0.0, 0.0])
    steps = 10
    for _ in range(steps):
        vehicle_action.step(action)

    ego_observation = construct_ego_observation(vehicle_model)
    ego_observation.observe(ego_lanelet, vehicle_action.vehicle, episode_length=20)
    observation = ego_observation.observation_dict
    if vehicle_model == 2:
        # Ego-related observations of KS model_type
        result = np.hstack((observation["v_ego"],
                            observation["a_ego"],
                            observation["steering_angle"],
                            observation["global_turn_rate"],
                            observation["remaining_steps"]))
    else:
        # Ego-related observations of PM model_type
        result = np.hstack((observation["v_ego"],
                            observation["a_ego"],
                            observation["remaining_steps"]))
    assert np.allclose(result, expected_output, rtol=0.01)


@pytest.mark.parametrize(
    ("vehicle_model", "action", "velocity", "steering_angle", "expected_output"),
    [
        # 0: VehicleModel.PM; 2: VehicleModel.KS
        (VehicleModel.PM, [0., 0.], [30., 0.], None, np.zeros(5)),
        (VehicleModel.PM, [0.5, 0.5], [30., 0.], None, np.zeros(5)),
        (VehicleModel.KS, [0., 0.], 30., 0.0, np.zeros(5)),
        (VehicleModel.KS, [0., 0.1], 30., 0.0, np.zeros(5)),
        (VehicleModel.KS, [0.02, 0.1], 30., 0.0, np.array([0, 0, 0, 0, 1])),
        (VehicleModel.KS, [0.05, 0.1], 30., 0.0, np.array([0, 0, 1, 1, 1])),
        (VehicleModel.KS, [0.2, 0.1], 30., 0.0, np.array([0, 1, 1, 1, 1])),
    ]
)
@functional
@unit_test
def test_check_friction_violation(vehicle_model, action, velocity, steering_angle, expected_output):
    vehicle_params = {
        "vehicle_type": 2,  # VehicleType.BMW_320i
        "vehicle_model": vehicle_model.value,  # 0: VehicleModel.PM; 2: VehicleModel.KS;
    }
    action_configs = {
        "action_type": "continuous", # discrete
        "action_base": "acceleration", # acceleration; jerk
        "long_steps": 5,
        "lat_steps": 5,

    }
    ego_observation = construct_ego_observation(vehicle_model.value)

    vehicle_action = ContinuousAction(vehicle_params, action_configs)
    if vehicle_params["vehicle_model"] == VehicleModel.KS.value:
        initial_state = CustomState(**dummy_state_ks, velocity=velocity, steering_angle=steering_angle, )
    else:  # VehicleModel.PM
        initial_state = CustomState(**dummy_state_pm, velocity=velocity[0], velocity_y=velocity[1])
    vehicle_action.reset(initial_state, dt=1.0)

    steps = 5
    result = np.zeros(steps)
    for i in range(steps):
        vehicle_action.step(action)
        ego_observation.observe(ego_lanelet, vehicle_action.vehicle, episode_length=20)
        result[i] = int(ego_observation.observation_dict["is_friction_violation"])
    assert np.all(result == expected_output)


@pytest.mark.parametrize(
    ("position", "polyline", "desired_orientation"),
    [
        (
                np.array([0, 0]),
                np.array([[0, 0], [1, 1], [0, 1]]) + 1,
                0.785398163
        ),
        (
                np.array([0, 0]),
                np.array([[100, -3], [0, 0], [-100, -3], [0, -103]]),
                3.111601,
        ),
        (
                np.array([0, 0]),
                np.array([[-7.3275, 12.5257],
                          [-7.5254, 9.1777],
                          [-11.278, 9.1652],
                          [-15.0305, 9.1526],
                          [-15.1272, 12.6073],
                          [-11.2273, 12.5665]]),
                -1.6298375
        ),
        (
                np.array([0, 0]),
                np.array([[-73.275, 12.5257],
                          [-75.254, 9.1777],
                          [-112.78, 9.1652],
                          [-150.305, 9.1526],
                          [-151.272, 12.6073],
                          [-112.273, 12.5665]]),
                -2.1046453
        ),
        (
                np.array([7, 12]),
                np.array([[-7.3275, 12.5257],
                          [-7.5254, 9.1777],
                          [-11.278, 9.1652],
                          [-15.0305, 9.1526],
                          [-15.1272, 12.6073],
                          [-11.2273, 12.5665]]),
                -1.6298375
        )
    ]
)
@functional
@unit_test
def test_get_orientation_of_polyline(position: np.ndarray, polyline: np.ndarray, desired_orientation: float):
    orientation = EgoObservation._get_orientation_of_polyline(position, polyline)
    assert np.isclose(orientation, desired_orientation)



# """
# Module containing the observation base class
# """
import logging
import warnings
from collections import OrderedDict
from typing import Union, Dict, List, Tuple, Set

import commonroad_dc.pycrcc as pycrcc
import gym
import numpy as np
from commonroad.geometry.shape import Polygon, Rectangle, Circle
from commonroad.planning.goal import GoalRegion
from commonroad.planning.planning_problem import PlanningProblem
from commonroad.scenario.lanelet import Lanelet, LaneletNetwork
from commonroad.scenario.obstacle import State, DynamicObstacle
from commonroad.scenario.scenario import Scenario
from commonroad.visualization.mp_renderer import MPRenderer, ZOrders
from commonroad.visualization.util import LineDataUnits
import commonroad_dc
from commonroad_dc.collision.collision_detection.pycrcc_collision_dispatch import create_collision_object, \
    create_collision_checker
from commonroad_dc.collision.trajectory_queries import trajectory_queries
from commonroad_route_planner.route_planner import RoutePlanner
from commonroad_dc.pycrccosy import CurvilinearCoordinateSystem

from commonroad_rl.gym_commonroad.action.vehicle import Vehicle
from commonroad_rl.gym_commonroad.observation.ego_observation import EgoObservation
from commonroad_rl.gym_commonroad.observation.goal_observation import GoalObservation
from commonroad_rl.gym_commonroad.observation.lanelet_network_observation import LaneletNetworkObservation
from commonroad_rl.gym_commonroad.observation.surrounding_observation import SurroundingObservation
from commonroad_rl.gym_commonroad.observation.traffic_sign_observation import TrafficSignObservation
from commonroad_rl.gym_commonroad.utils.navigator import Navigator
from commonroad_rl.gym_commonroad.utils.conflict_zone import ConflictZone

LOGGER = logging.getLogger(__name__)


class ObservationCollector:
    """
    This class is a wrapper for individual observation classes. It serves as an access point for the observations.
    Currently, this class supports the GoalObservation, SurroundingObservation, LaneletNetworkObservation,
    EgoObservation.

    :param configs: dictionary with the parameters of the configuration
        should include a individual dictionary for the individual observation classes

    Examples::

        observation_collector = ObservationCollector(configs)
        # build gym observation space
        observation_space = observation_collector._build_observation_space()
        # reset and observe
        observation_collector.reset()
        observation = observation_collector.observe(ego_vehicle)
    """

    def __init__(self, configs: Dict):
        self._scenario: Scenario = None
        self._goal_region: GoalRegion = None
        self._planning_problem: PlanningProblem = None
        self.observation_space_size: int = None
        self._flatten_observation = configs.get("flatten_observation")
        self._max_lane_merge_range = configs.get("max_lane_merge_range")
        self.observation_dict = OrderedDict()

        self.ego_observation = EgoObservation(configs)
        self.goal_observation = GoalObservation(configs)
        self.surrounding_observation = SurroundingObservation(configs)
        self.lanelet_network_observation = LaneletNetworkObservation(configs)
        self.traffic_sign_observation = TrafficSignObservation(configs)

        self._road_edge = dict()
        self._cache_lanelet_polygons_accel_struct = dict()
        self._cache_lanelet_polygons_sg_accel_struct = dict()
        self._cache_scenario_ref_path_dict = dict()
        self._connected_lanelet_dict: Dict[int: Set[int]] = dict()
        self._cache_collision_checker_templates: Dict[str: pycrcc.CollisionChecker] = dict()

        self.observation_space = self._build_observation_space()

        self.navigator: Navigator = None
        self._use_cache_navigator = configs.get("cache_navigators", False)
        self._continous_collision_check = configs["action_configs"].get("continuous_collision_checking", True)
        self._cache_navigator = dict()
        self.episode_length = None
        self._enable_conflit_zone = self.surrounding_observation.observe_intersection_velocities \
                                    or self.surrounding_observation.observe_intersection_distances \
                                    or self.surrounding_observation.observe_ego_distance_intersection
        if self._enable_conflit_zone:
            self.conflict_zone = ConflictZone()
        else:
            self.conflict_zone = None

    def _build_observation_space(self) -> Union[gym.spaces.Box, gym.spaces.Dict]:
        """
        builds the observation space dictionary

        :return: the function returns an OrderedDict with the observation spaces of each observation as an entry
        """
        observation_space_dict = OrderedDict()
        observation_space_dict.update(self.goal_observation.build_observation_space())
        observation_space_dict.update(self.ego_observation.build_observation_space())
        observation_space_dict.update(self.surrounding_observation.build_observation_space())
        observation_space_dict.update(self.lanelet_network_observation.build_observation_space())
        observation_space_dict.update(self.traffic_sign_observation.build_observation_space())

        self.observation_space_dict = observation_space_dict
        if self._flatten_observation:
            lower_bounds, upper_bounds = np.array([]), np.array([])
            for space in observation_space_dict.values():
                lower_bounds = np.concatenate((lower_bounds, space.low))
                upper_bounds = np.concatenate((upper_bounds, space.high))
            self.observation_space_size = lower_bounds.shape[0]
            observation_space = gym.spaces.Box(low=lower_bounds, high=upper_bounds, dtype=np.float64)
            LOGGER.debug(f"Size of flattened observation space: {self.observation_space_size}")
        else:
            observation_space = gym.spaces.Dict(self.observation_space_dict)
            LOGGER.debug(f"Length of dictionary observation space: {len(self.observation_space_dict)}")

        return observation_space

    def reset(self, scenario: Scenario, planning_problem: PlanningProblem, reset_config: dict, benchmark_id: str,
              clone_collision_checker: bool = True):

        self.time_step = 0
        self._scenario = scenario
        self._planning_problem = planning_problem
        if self._planning_problem is not None:
            self._goal_region: GoalRegion = planning_problem.goal
            self.episode_length = max(s.time_step.end for s in self._goal_region.state_list)
        else:
            self._goal_region = None
            self.episode_length = None
        self._benchmark_id = benchmark_id
        self._update_collision_checker(clone_collision_checker=clone_collision_checker)

        self.ego_lanelet_id = None
        self.ego_lanelet = None

        self.navigator = None

        self.goal_observation.observation_history_dict = dict()

        # dictionary is not set anywhere
        if self._cache_scenario_ref_path_dict.get(str(self._scenario.scenario_id), None) is None:
            self._cache_scenario_ref_path_dict[str(self._scenario.scenario_id)] = dict()

        self._connected_lanelet_dict: Dict[int: Set[int]] = reset_config["connected_lanelet_dict"]

        self._road_edge = {
            "left_road_edge_lanelet_id_dict": reset_config["left_road_edge_lanelet_id_dict"],
            "right_road_edge_lanelet_id_dict": reset_config["right_road_edge_lanelet_id_dict"],
            "left_road_edge_dict": reset_config["left_road_edge_dict"],
            "right_road_edge_dict": reset_config["right_road_edge_dict"],
            "boundary_collision_object": reset_config["boundary_collision_object"], }
        #TODO: make this usable for more than the inD data set
        if self._enable_conflit_zone and scenario.scenario_id.__str__().startswith('DEU_AAH'):
            self.conflict_zone.reset(scenario)

    @property
    def road_edge(self):
        return self._road_edge

    @staticmethod
    def compute_convex_hull_circle(radius, previous_position, current_position) -> pycrcc.RectOBB:
        """ Compute obb based on last and current position to
            approximate the area covered by the collision circle between
            the last and current timestep.
        """
        position = (current_position + previous_position) / 2.0
        direction = current_position - previous_position
        direction_length = np.linalg.norm(direction)
        d_normed = direction / direction_length
        orientation = np.arctan2(d_normed[1], d_normed[0])

        return pycrcc.RectOBB(direction_length / 2 + radius, radius, orientation, position[0], position[1])

    @staticmethod
    def create_convex_hull_collision_circle(dynamic_obstacle: DynamicObstacle):
        assert isinstance(dynamic_obstacle.obstacle_shape, Circle)

        initial_time_step = dynamic_obstacle.initial_state.time_step
        tvo = pycrcc.TimeVariantCollisionObject(initial_time_step)
        if dynamic_obstacle.prediction is not None:
            for time_step in range(initial_time_step, dynamic_obstacle.prediction.final_time_step):
                previous_state = dynamic_obstacle.state_at_time(time_step)
                state = dynamic_obstacle.state_at_time(time_step + 1)
                convex_obb = ObservationCollector.compute_convex_hull_circle(
                    dynamic_obstacle.obstacle_shape.radius, previous_state.position, state.position)
                tvo.append_obstacle(convex_obb)
        else: # SUMO scenario
            tvo.append_obstacle(create_collision_object(dynamic_obstacle))

        return tvo

    @staticmethod
    def create_collision_checker_scenario(scenario: Scenario, params=None, collision_object_func=None, continous_collision_check=True):
        if not continous_collision_check:
            return create_collision_checker(scenario)
        cc = pycrcc.CollisionChecker()
        for co in scenario.dynamic_obstacles:
            if isinstance(co.obstacle_shape, Rectangle):
                collision_object = create_collision_object(co, params, collision_object_func)
                if co.prediction is not None:
                # TODO: remove if when https://gitlab.lrz.de/cps/commonroad-drivability-checker/-/issues/16 is fixed
                    collision_object, err = trajectory_queries.trajectory_preprocess_obb_sum(collision_object)
                    if err:
                        raise Exception("<ObservationCollector.create_collision_checker_scenario> Error create convex hull")
            elif isinstance(co.obstacle_shape, Circle):
                collision_object = ObservationCollector.create_convex_hull_collision_circle(co)
            else:
                raise NotImplementedError(f"Unsupported shape for convex hull collision object: {co.obstacle_shape}")
            cc.add_collision_object(collision_object)

        shape_group = pycrcc.ShapeGroup()
        for co in scenario.static_obstacles:
            collision_object = commonroad_dc.collision.collision_detection.pycrcc_collision_dispatch. \
                create_collision_object(co, params, collision_object_func)
            if isinstance(collision_object, pycrcc.ShapeGroup):
                for shape in collision_object.unpack():
                    shape_group.add_shape(shape)
            else:
                shape_group.add_shape(collision_object)
        cc.add_collision_object(shape_group)
        return cc

    def _update_collision_checker(self, collision_checker=None, clone_collision_checker: bool = True):

        if collision_checker is None:
            if clone_collision_checker:
                full_reset = False
                cc_template = self._cache_collision_checker_templates.get(self._benchmark_id, None)
                if cc_template is None:
                    cc_template = self.create_collision_checker_scenario(
                        self._scenario, continous_collision_check=self._continous_collision_check)
                    if not full_reset:
                        self._cache_collision_checker_templates[self._benchmark_id] = cc_template
                self._collision_checker = cc_template.clone()
            else:
                self._collision_checker = self.create_collision_checker_scenario(
                    self._scenario, continous_collision_check=self._continous_collision_check)
        else:
            self._collision_checker = collision_checker

    def get_collision_checker(self):
        return self._collision_checker

    def observe(self, ego_vehicle: Vehicle) -> Union[np.array, Dict]:
        # initialize observation_dict
        observation_dict = OrderedDict()
        self._ego_vehicle = ego_vehicle
        self._ego_state = ego_vehicle.state
        self.time_step = self._ego_state.time_step
        # TODO: only valid for car env
        self._update_ego_lanelet_and_local_ccosy()
        self._create_navigator()

        # calculate all observations
        observation_dict_ego = self.ego_observation.observe(self.ego_lanelet, ego_vehicle, self.episode_length)
        observation_dict_goal = self.goal_observation.observe(self._ego_state, goal=self._goal_region,
                                                              scenario=self._scenario,
                                                              planning_problem=self._planning_problem,
                                                              ego_lanelet_ids=self._ego_lanelet_ids,
                                                              navigator=self.navigator,
                                                              episode_length=self.episode_length,
                                                              local_ccosy=self.local_ccosy)

        observation_dict_surrounding, ego_vehicle_lat_position = self.surrounding_observation.observe(self._scenario,
                                                                                                      ego_vehicle,
                                                                                                      self.time_step,
                                                                                                      self._connected_lanelet_dict,
                                                                                                      self.ego_lanelet,
                                                                                                      self._collision_checker,
                                                                                                      self.local_ccosy,
                                                                                                      self.conflict_zone,
                                                                                                      self._ego_lanelet_ids)

        observation_dict_lanelet = self.lanelet_network_observation.observe(self._scenario, ego_vehicle,
                                                                            self.ego_lanelet, self.road_edge,
                                                                            self.local_ccosy, self.navigator)
        observation_dict_traffic_sign = self.traffic_sign_observation.observe(self._scenario, ego_vehicle,
                                                                              self.ego_lanelet, self.local_ccosy)

        observation_dict.update(observation_dict_ego)
        observation_dict.update(observation_dict_surrounding)
        observation_dict.update(observation_dict_lanelet)
        observation_dict.update(observation_dict_goal)
        observation_dict.update(observation_dict_traffic_sign)

        assert len(list(observation_dict.keys())) == len(list(self.observation_space_dict.keys()))
        self.observation_dict = OrderedDict((k, observation_dict[k]) for k in self.observation_space_dict.keys())
        assert list(self.observation_dict.keys()) == list(self.observation_space_dict.keys())

        if self._flatten_observation:
            observation_vector = np.zeros(self.observation_space.shape)
            index = 0
            for k in self.observation_dict.keys():
                size = np.prod(self.observation_dict[k].shape)
                observation_vector[index: index + size] = self.observation_dict[k].flat
                index += size
            return observation_vector
        else:
            return self.observation_dict

    def _update_ego_lanelet_and_local_ccosy(self):
        # update ego lanelet
        lanelet_polygons, lanelet_polygons_sg = self._get_lanelet_polygons(str(self._scenario.scenario_id))
        self._ego_lanelet_ids = self.sorted_lanelets_by_state(self._scenario, self._ego_state, lanelet_polygons,
                                                              lanelet_polygons_sg)
        if len(self._ego_lanelet_ids) == 0:
            ego_lanelet_id = self.ego_lanelet_id
        else:
            self.ego_lanelet_id = ego_lanelet_id = self._ego_lanelet_ids[0]

        self.ego_lanelet = self._scenario.lanelet_network.find_lanelet_by_id(ego_lanelet_id)
        ref_path_dict = self._cache_scenario_ref_path_dict[str(self._scenario.scenario_id)]

        # update local ccosy
        self.local_ccosy, self._local_merged_lanelet = self.get_local_curvi_cosy(self._scenario.lanelet_network,
                                                                                 ego_lanelet_id, ref_path_dict,
                                                                                 self._max_lane_merge_range)

    def render(self, render_configs: Dict, render: MPRenderer):
        self.lanelet_network_observation.draw(render_configs, render,
                                              ego_vehicle=self._ego_vehicle, road_edge=self.road_edge,
                                              ego_lanelet=self.ego_lanelet, navigator=self.navigator)
        self.goal_observation.draw(render_configs, render, self.navigator)
        self.surrounding_observation.draw(render_configs, render)
        # Plot ccosys
        if render_configs["render_local_ccosy"]:
            # draw_params = ParamServer(
            #     {"lanelet":
            #         {"unique_colors": False,
            #         "draw_stop_line": False, "stop_line_color": "#ffffff", "draw_line_markings": False,
            #         "draw_left_bound": False, "draw_right_bound": False, "draw_center_bound": False,
            #         "draw_border_vertices": False, "draw_start_and_direction": False, "show_label": False,
            #         "draw_linewidth": 0.5, "fill_lanelet": True, "facecolor": "wheat"},
            #      "traffic_sign": {"draw_traffic_signs": False, "show_traffic_signs": "all", "show_label": True}
            #      }
            # )
            # render.draw_list(self._local_merged_lanelet, draw_params=draw_params)
            # # TODO: temporary fix for missing draw method in Lanelet, remove after fixed in cr-io
            # from commonroad.scenario.lanelet import LaneletNetwork
            # LaneletNetwork.create_from_lanelet_list([self._local_merged_lanelet]).draw(render, draw_params=draw_params)
            # self._local_merged_lanelet.draw(render, draw_params=draw_params)
            reference_path_local_ccosy = np.array(self.local_ccosy.reference_path())
            line = LineDataUnits(reference_path_local_ccosy[:, 0], reference_path_local_ccosy[:, 1],
                                 zorder=ZOrders.LANELET_LABEL, markersize=1., color="green",
                                 marker="x", label="Local CCOSy")
            render.dynamic_artists.append(line)

    def _get_lanelet_polygons(self, meta_scenario_id) -> Tuple[List[Tuple[int, Polygon]], pycrcc.ShapeGroup]:
        """
        returns lanelet_polygons and the shape group of the polygons
        # TODO: fix meta_scenario_id
        :param meta_scenario_id: id of the current meta scenario
        :return:
            lanelet_polygons: List of tuples for each lanelet the tuple contains the lanelet id and the corresponding
                polygon
            lanelet+polygons_sg: ShapeGroup of all the polygons of all the lanelets
        """
        # TODO: pre store polygon for all maps
        lanelet_polygons = self._cache_lanelet_polygons_accel_struct.get(meta_scenario_id, None)
        lanelet_polygons_sg = self._cache_lanelet_polygons_sg_accel_struct.get(meta_scenario_id, None)

        if lanelet_polygons is None:
            lanelet_polygons = [(lanelet.lanelet_id, lanelet.convert_to_polygon()) for lanelet in
                                self._scenario.lanelet_network.lanelets]
            lanelet_polygons_sg = pycrcc.ShapeGroup()
            for l_id, poly in lanelet_polygons:
                lanelet_polygons_sg.add_shape(create_collision_object(poly))
            self._cache_lanelet_polygons_sg_accel_struct[meta_scenario_id] = lanelet_polygons_sg
            self._cache_lanelet_polygons_accel_struct[meta_scenario_id] = lanelet_polygons
        return lanelet_polygons, lanelet_polygons_sg

    @staticmethod
    def sorted_lanelets_by_state(scenario: Scenario, state: State, lanelet_polygons: list,
                                 lanelet_polygons_sg: pycrcc.ShapeGroup) -> List[int]:
        """
        Returns the sorted list of lanelet ids which correspond to a given state

        :param scenario: The scenario to be used
        :param state: The state which lanelets ids are searched
        :param lanelet_polygons: The polygons of the lanelets
        :param lanelet_polygons_sg: Thy pycrcc polygons of the lanelets
        :return: The list of lanelet ids sorted by relative orientations, the nearest lanelet is the first elements
        """
        return Navigator.sorted_lanelet_ids(
            ObservationCollector._related_lanelets_by_state(state, lanelet_polygons, lanelet_polygons_sg),
            state.orientation if hasattr(state, "orientation") else np.arctan2(state.velocity_y, state.velocity),
            state.position, scenario, )

    @staticmethod
    def _related_lanelets_by_state(state: State, lanelet_polygons: List[Tuple[int, Polygon]],
                                   lanelet_polygons_sg: pycrcc.ShapeGroup) -> List[int]:
        """
        Get the lanelet of a state.

        :param state: The state to which the related lanelets should be found
        :param lanelet_polygons: The polygons of the lanelets
        :param lanelet_polygons_sg: The pycrcc polygons of the lanelets
        :return: The list of lanelet ids
        """
        # output list
        res = list()

        # look at each lanelet
        point_list = [state.position]

        point_sg = pycrcc.ShapeGroup()
        for el in point_list:
            point_sg.add_shape(pycrcc.Point(el[0], el[1]))

        lanelet_polygon_ids = point_sg.overlap_map(lanelet_polygons_sg)

        for lanelet_id_list in lanelet_polygon_ids.values():
            for lanelet_id in lanelet_id_list:
                res.append(lanelet_polygons[lanelet_id][0])

        return res

    @staticmethod
    def get_local_curvi_cosy(lanelet_network: LaneletNetwork, ego_vehicle_lanelet_id: int,
                             ref_path_dict: Dict[str, Tuple[np.ndarray or None, Lanelet or None]],
                             max_lane_merge_range: float) -> Tuple[CurvilinearCoordinateSystem, Lanelet]:
        """
        At every time step, update the local curvilinear coordinate system from the dict.

        :param lanelet_network: The lanelet network
        :param ego_vehicle_lanelet_id: The lanelet id where the ego vehicle is on
        :param ref_path_dict: The dictionary of the reference path, contains the paths by the starting lanelet ids
        :param max_lane_merge_range: Maximum range of lanes to be merged
        :return: Curvilinear coordinate system of the merged lanelets
        """
        if ref_path_dict is None:
            ref_path_dict = dict()

        ref_path, ref_merged_lanelet, curvi_cosy = ref_path_dict.get(ego_vehicle_lanelet_id, (None, None, None))

        if curvi_cosy is None:
            for lanelet in lanelet_network.lanelets:  # iterate in all lanelet in this scenario
                if lanelet.lanelet_id == ego_vehicle_lanelet_id and (
                        not lanelet.predecessor and not lanelet.successor):  # the lanelet is a lane itself
                    ref_path = lanelet.center_vertices
                    ref_merged_lanelet = lanelet
                elif not lanelet.predecessor:  # the lanelet is the start of a lane, the lane can be created from here
                    # TODO: cache merged lanelets in pickle or dict
                    merged_lanelet_list, sub_lanelet_ids_list = Lanelet.all_lanelets_by_merging_successors_from_lanelet(
                        lanelet, lanelet_network, max_lane_merge_range)
                    for merged_lanelet, sub_lanelet_ids in zip(merged_lanelet_list, sub_lanelet_ids_list):
                        if ego_vehicle_lanelet_id in sub_lanelet_ids:
                            ref_path = merged_lanelet.center_vertices
                            ref_merged_lanelet = merged_lanelet
                            break
            # TODO: Idea, the reference path dict could be updated
            #  on all successor of the current lanelet for optimization
            curvi_cosy = Navigator.create_coordinate_system_from_polyline(ref_path)
            ref_path_dict[ego_vehicle_lanelet_id] = (ref_path, ref_merged_lanelet, curvi_cosy)

        return curvi_cosy, ref_merged_lanelet

    def _create_navigator(self) -> Navigator:
        """
        creates and stores the Navigator of the current scenario

        """

        if not self.navigator:

            # Check cache first
            if self._use_cache_navigator:
                key = ObservationCollector.get_navigator_cache_key(self._scenario, self._planning_problem)
                if key in self._cache_navigator:
                    self.navigator = self._cache_navigator[key]
                    return self._cache_navigator[key]

            # not found in cache --> create new navigator
            route_planner = RoutePlanner(self._scenario, self._planning_problem,
                                         backend=RoutePlanner.Backend.NETWORKX_REVERSED, log_to_console=False, )

            route_candidates = route_planner.plan_routes()
            route = route_candidates.retrieve_best_route_by_orientation()

            navigator = Navigator(route)

            if self._use_cache_navigator:
                self._cache_navigator[key] = navigator

            self.navigator = navigator
            return navigator

    @staticmethod
    def get_navigator_cache_key(scenario: Scenario, planning_problem: PlanningProblem):
        return str(scenario.scenario_id) + '#' + str(planning_problem.planning_problem_id)


if __name__ == "__main__":
    import yaml
    from commonroad_rl.gym_commonroad.constants import PATH_PARAMS

    config_file = PATH_PARAMS["configs"]["commonroad-v1"]
    with open(config_file, "r") as config_file:
        config = yaml.safe_load(config_file)
    configs = config["env_configs"]
    observation_collector = ObservationCollector(configs)
    print(observation_collector)



"""
Module for CommonRoad vectorized environment used in Stable Baselines
"""
import time
from typing import Callable

import numpy as np
from gym import Env
from stable_baselines.common.vec_env import DummyVecEnv


class CommonRoadVecEnv(DummyVecEnv):
    def __init__(self, env_fns):
        super().__init__(env_fns)
        self.on_reset = None
        self.start_times = np.array([])

    def set_on_reset(self, on_reset_callback: Callable[[Env, float], None]):
        self.on_reset = on_reset_callback

    def reset(self, **kwargs):
        self.start_times = np.array([time.time()] * self.num_envs)
        # copied from DummyVecEnv to enable reset with kwargs
        for env_idx in range(self.num_envs):
            obs = self.envs[env_idx].reset(**kwargs)
            self._save_obs(env_idx, obs)
        return self._obs_from_buf()

    def step_wait(self):
        out_of_scenarios = False
        for env_idx in range(self.num_envs):
            (obs, self.buf_rews[env_idx], self.buf_dones[env_idx], self.buf_infos[env_idx],) = self.envs[env_idx].step(
                np.squeeze(self.actions[env_idx]))
            if self.buf_dones[env_idx]:
                # save final observation where user can get it, then reset
                self.buf_infos[env_idx]["terminal_observation"] = obs

                # Callback
                elapsed_time = time.time() - self.start_times[env_idx]
                self.on_reset(self.envs[env_idx], elapsed_time)
                self.start_times[env_idx] = time.time()

                # If one of the environments doesn't have anymore scenarios it will throw an Exception on reset()
                try:
                    obs = self.envs[env_idx].reset()
                except IndexError:
                    out_of_scenarios = True
            self._save_obs(env_idx, obs)
            self.buf_infos[env_idx]["out_of_scenarios"] = out_of_scenarios
        return self._obs_from_buf(), np.copy(self.buf_rews), np.copy(self.buf_dones), self.buf_infos.copy()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""commonroad_rl setup file."""

import sys

try:
    from setuptools import setup, find_packages
except ImportError:
    print("Please install or upgrade setuptools or pip to continue")
    sys.exit(1)

setup(
    name="commonroad-rl",
    version="2023.1.3",
    packages=find_packages(
        exclude=["tests", "utils_run"]
    ),
    package_data={"": ["*.xml", "*.pickle", "*.yaml"]},
    description="Tools for applying reinforcement learning on commonroad scenarios.",
    long_description=open("README.md", encoding="utf8").read(),
    long_description_content_type="text/markdown",
    test_suite="commonroad_rl.tests",
    keywords="autonomous automated vehicles driving motion planning".split(),
    url="https://commonroad.in.tum.de/",
    install_requires=[
        "aenum==3.1.11",
        "numpy",
        "scipy",
        "networkx",
        "triangle",
        "optuna",
        "pytest",
        "importlib-metadata==4.13.0",
        "Polygon3==3.0.9.1",
        "pybind11==2.10.1",
        "tensorflow==1.15.0",
        "shapely==1.8.5",
        "setuptools==50.3.2",
        "gym==0.21.0",
        "stable_baselines",
        "commonroad-io==2022.3",
        "commonroad-vehicle-models==3.0.2",
        "commonroad-drivability-checker==2022.2.1",
        "commonroad-route-planner==2022.3",
        "SALib==1.3.13"
    ],
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Operating System :: POSIX :: Linux"
    ],
    include_package_data=True
)



"""
A short tool to assess the number and time spanned by the files resulting from the conversion
Pass the path to the problem pickles as first command line argument.
"""

__author__ = "Niels Mündler"
__copyright__ = ""
__credits__ = [""]
__version__ = "1.0"
__maintainer__ = "Xiao Wang"
__email__ = "xiao.wang@tum.de"
__status__ = "Released"

from pathlib import Path
import pickle
from argparse import ArgumentParser


def get_parser():
    parser = ArgumentParser()
    parser.add_argument(
        "pickles_path",
        default="pickles/problem",
        help="Path to pickled problem scenarios",
    )
    parser.add_argument(
        "-t",
        "--time_step",
        type=float,
        default=0.04,
        help="Time step length of the pickled scenario. Default is 0.04 seconds per step = 30 fps at the original recording",
        dest="time_step",
    )
    return parser


if __name__ == "__main__":
    args = get_parser().parse_args()
    pickles_path = Path(args.pickle_path)
    if not pickles_path.is_dir():
        print("Error: path {} is not a directory".format(pickles_path))
        exit(-1)

    overall_time = 0
    read_scenarios = 0
    for problem_path in pickles_path.iterdir():
        print(problem_path)
        try:
            with problem_path.open("rb") as file:
                pickled_problem = pickle.load(file)
                planning_problem_set = pickled_problem["planning_problem_set"]
                read_scenarios += 1
                for p in planning_problem_set.planning_problem_dict.values():
                    overall_time += max(s.time_step.end for s in p.goal.state_list)
        except Exception as e:
            print(e)
            pass
    print(f"Read {read_scenarios} scenarios")
    print(f"Overall training timesteps: {overall_time}")
    seconds = args.time_step * overall_time
    print(f"This is equivalent to {seconds} seconds")
    hours = (seconds / 60) / 60
    print(f"This is equivalent to {hours} hours")



from collections import defaultdict, OrderedDict
from typing import Union, Dict, List, Tuple

import gym
import numpy as np
from commonroad.scenario.lanelet import Lanelet
from commonroad.scenario.scenario import Scenario
from commonroad.scenario.traffic_sign import TrafficSign, TrafficSignIDGermany
from commonroad_rl.gym_commonroad.action.vehicle import Vehicle
from commonroad_rl.gym_commonroad.observation.observation import Observation
from numpy import ndarray
from commonroad_dc.pycrccosy import CurvilinearCoordinateSystem
from shapely.geometry import Point, LineString


class TrafficSignObservation(Observation):
    def __init__(self, configs: Dict, configs_name: str = "traffic_sign_configs"):
        # Read configs
        configs = configs[configs_name]
        self.observe_stop_sign: bool = configs.get("observe_stop_sign")
        self.observe_yield_sign: bool = configs.get("observe_yield_sign")
        self.observe_priority_sign: bool = configs.get("observe_priority_sign")
        self.observe_right_of_way_sign: bool = configs.get("observe_right_of_way_sign")

        self.observation_dict = OrderedDict()
        self.observation_history_dict = defaultdict(list)
        self.traffic_sign_id_list = []
        self.traffic_sign_id_list_successor = []

    def build_observation_space(self) -> OrderedDict:
        observation_space_dict = OrderedDict()

        if self.observe_stop_sign:
            observation_space_dict["stop_sign"] = gym.spaces.Box(0, 1, (1,), dtype=np.int8)
            observation_space_dict["stop_sign_distance_long"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)

        if self.observe_yield_sign:
            observation_space_dict["yield_sign"] = gym.spaces.Box(0, 1, (1,), dtype=np.int8)
            observation_space_dict["yield_sign_distance_long"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)

        if self.observe_priority_sign:
            observation_space_dict["priority_sign"] = gym.spaces.Box(0, 1, (1,), dtype=np.int8)
            observation_space_dict["priority_sign_distance_long"] = gym.spaces.Box(-np.inf, np.inf, (1,),
                                                                                   dtype=np.float32)
        if self.observe_right_of_way_sign:
            observation_space_dict["right_of_way_sign"] = gym.spaces.Box(0, 1, (1,), dtype=np.int8)
            observation_space_dict["right_of_way_sign_distance_long"] = gym.spaces.Box(-np.inf, np.inf, (1,),
                                                                                       dtype=np.float32)
        return observation_space_dict

    def observe(self, scenario: Scenario, ego_vehicle: Vehicle, ego_vehicle_lanelet: Lanelet,
                local_ccosy: Union[None, CurvilinearCoordinateSystem] = None) -> Union[ndarray, Dict]:
        """
        Create traffic sign observation for given state in an environment.
        """
        self.traffic_sign_id_list, self.traffic_sign_id_list_successor = \
            self._get_traffic_signs_on_lanelet_and_successor(scenario, ego_vehicle_lanelet)

        if self.observe_stop_sign:
            self._observe_traffic_sign("stop_sign", TrafficSignIDGermany.STOP, ego_vehicle,
                                       ego_vehicle_lanelet, scenario, True)

        if self.observe_yield_sign:
            self._observe_traffic_sign("yield_sign", TrafficSignIDGermany.YIELD, ego_vehicle,
                                       ego_vehicle_lanelet, scenario, True)

        if self.observe_priority_sign:
            self._observe_traffic_sign("priority_sign", TrafficSignIDGermany.PRIORITY, ego_vehicle,
                                       ego_vehicle_lanelet, scenario)

        if self.observe_right_of_way_sign:
            self._observe_traffic_sign("right_of_way_sign", TrafficSignIDGermany.RIGHT_OF_WAY, ego_vehicle,
                                       ego_vehicle_lanelet, scenario)

        return self.observation_dict

    def _observe_traffic_sign(self, sign_entry: str, sign_type: TrafficSignIDGermany, ego_vehicle: Vehicle,
                              ego_vehicle_lanelet: Lanelet, scenario: Scenario, successor=False):
        """
        finds traffic sign in the current or successor lanelets and distance to the sign (-1 if not present)
        :param sign_entry: sign entry name in the observation dictionary
        :param sign_type: traffic sign id enum
        :param ego_vehicle: the ego vehicle
        :param ego_vehicle_lanelet: the lanelet of the ego vehicle
        :param scenario: scenario with traffic signs
        """
        sign_distance_entry = f"{sign_entry}_distance_long"

        self.observation_dict[sign_entry] = np.array([0])
        self.observation_dict[sign_distance_entry] = np.array([-1])

        search_list = self.traffic_sign_id_list_successor if successor else self.traffic_sign_id_list
        # add traffic signs on current lanelet and successor lanelets
        for sign_id in search_list:
            sign = scenario.lanelet_network.find_traffic_sign_by_id(sign_id)
            if sign.traffic_sign_elements[0].traffic_sign_element_id == sign_type:
                self.observation_dict[sign_entry] = np.array([1])

                # calculate distance from ego vehicle head to the forward boundary of current lanelet
                forward_distance = self.get_distance_to_traffic_sign_long(scenario, ego_vehicle,
                                                                          ego_vehicle_lanelet, sign, successor)
                self.observation_dict[sign_distance_entry] = np.array([forward_distance])
                return

    @staticmethod
    def get_distance_to_traffic_sign_long(scenario: Scenario, ego_vehicle: Vehicle,
                                          ego_vehicle_lanelet: Lanelet,
                                          sign: TrafficSign, successor=False) -> float:
        """
        get distance from the head of ego vehicle to forward boundary of current lanelet
        :param scenario: current scenario
        :param ego_vehicle: the ego vehicle
        :param ego_vehicle_lanelet: lanelet of ego vehicle
        :param sign: observed traffic sign on current or successor lanelet
        :param successor: whether including traffic sign on the successor lanelet, True given stop and yield sign

        :return: distance from ego vehicle to forward boundary, returns -1.0 if the sign could not be found
        """
        # Check for stop line in ego and successor lanelets
        stop_line_present = \
            ego_vehicle_lanelet.stop_line and ego_vehicle_lanelet.stop_line.traffic_sign_ref == sign.traffic_sign_id
        if stop_line_present:
            stop_line = ego_vehicle_lanelet.stop_line
            forward_boundary = LineString((stop_line.start, stop_line.end))
        elif successor:
            for suc_id in ego_vehicle_lanelet.successor:
                suc_lanelet = scenario.lanelet_network.find_lanelet_by_id(suc_id)
                stop_line_present = \
                    suc_lanelet.stop_line and suc_lanelet.stop_line.traffic_sign_ref == sign.traffic_sign_id
                if stop_line_present:
                    stop_line = suc_lanelet.stop_line
                    forward_boundary = LineString((stop_line.start, stop_line.end))
                    break

        if not stop_line_present:
            if sign.traffic_sign_id in ego_vehicle_lanelet.traffic_signs:
                forward_boundary = LineString(
                    (ego_vehicle_lanelet.right_vertices[-1], ego_vehicle_lanelet.left_vertices[-1]))
            elif successor:
                for suc_id in ego_vehicle_lanelet.successor:
                    suc_lanelet = scenario.lanelet_network.find_lanelet_by_id(suc_id)
                    if sign.traffic_sign_id in suc_lanelet.traffic_signs:
                        forward_boundary = LineString((suc_lanelet.right_vertices[-1], suc_lanelet.left_vertices[-1]))
                        break
                # Couldn't find traffic sign, something went wrong (possibly mismatched successor flags when searching
                # for the sign and calculating the distance)
                try:
                    forward_boundary
                except NameError:
                    return -1.0
            # Traffic sign can't be found with current settings
            else:
                return -1.0

        # Calculate the actual distance
        ego_vehicle_point = Point(ego_vehicle.state.position)
        nearest_point = forward_boundary.interpolate(forward_boundary.project(ego_vehicle_point))
        forward_distance = nearest_point.distance(ego_vehicle_point)

        return abs(forward_distance)

    def _get_traffic_signs_on_lanelet_and_successor(self, scenario: Scenario, ego_vehicle_lanelet: Lanelet) \
            -> Tuple[List[int], List[int]]:
        """
        get traffic signs which located on current and successor lanelets
        :param scenario: the current scenario
        :param ego_vehicle_lanelet: lanelet of ego vehicle

        :return: list of traffic signs on current lanelet and a list of traffic signs on current and successor lanelets
        """
        traffic_sign_list = []
        traffic_sign_list += list(ego_vehicle_lanelet.traffic_signs)
        traffic_sign_list_successor = list.copy(traffic_sign_list)
        for successor_id in ego_vehicle_lanelet.successor:
            successor_lanelet = scenario.lanelet_network.find_lanelet_by_id(successor_id)
            if len(successor_lanelet.traffic_signs) != 0:
                traffic_sign_list_successor += list(successor_lanelet.traffic_signs)
        return traffic_sign_list, traffic_sign_list_successor

    def draw(self):
        pass


if __name__ == "__main__":
    import yaml
    from commonroad_rl.gym_commonroad.constants import PATH_PARAMS

    config_file = PATH_PARAMS["configs"]["commonroad-v1"]
    with open(config_file, "r") as config_file:
        config = yaml.safe_load(config_file)
    configs = config["env_configs"]
    traffic_sign_observation = TrafficSignObservation(configs)
    print(traffic_sign_observation)
